##################################
##### Tests for taeg build #####
##################################

import unittest
import os
import shutil
import testing.postgresql

from psycopg2 import connect, extensions
from psycopg2.errors import DuplicateTable

from taeg.argparser import get_parser
from taeg.service.build import write_assignment_info, write_class_info
from taeg.service.build import main as build
from taeg.service.create import main as create

from .. import TestCase

TEST_FILES_PATH = "test/test_service/test-build/"

parser = get_parser()

class TestBuild(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        cls.postgresql = testing.postgresql.Postgresql()
        cls.conn = connect(**cls.postgresql.dsn())
        cls.conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)

        args = parser.parse_args(["service", "create"])
        args.func = create
        args.func(args, conn=cls.conn, close_conn=False)

        args = parser.parse_args(["service", "build", TEST_FILES_PATH, "-q"])
        args.func = build
        args.func(args, conn=cls.conn, close_conn=False) # Function has built-in assert statement for error-checking

        cls.cursor = cls.conn.cursor()
        cls.cursor.execute("""INSERT INTO classes (class_id, class_name)
                                VALUES ('1234', 'dummy name')""")
        cls.cursor.execute("""INSERT INTO classes (class_id, class_name)
                                VALUES ('2345', 'dummy name')""")
        cls.cursor.execute("""INSERT INTO assignments (assignment_id, class_id, assignment_name)
                                VALUES ('dummyid', '1234', 'dummy name')""")


    def test_write_class_info(self):
        class_id = write_class_info("test_class" ,"test class", self.conn)
        self.cursor.execute("""SELECT * FROM classes
                                WHERE class_id = %s""", (class_id, ))
        assert self.cursor.rowcount == 1, "Couldn't find inserted class"


    def test_write_assignment_info(self):
        write_assignment_info("dummy_aid", 1234, "dummy class name", None, self.conn)
        write_assignment_info("dummy_aid2", 2345, "dummy class name 2", None, self.conn)
        self.cursor.execute("""SELECT * FROM classes WHERE class_id = '2345'""")
        assert self.cursor.rowcount == 1, "Couldn't find inserted assignment information for dummy_aid2"
    

    @classmethod
    def tearDownClass(cls):
        cls.cursor.close()
        cls.conn.close()
        cls.postgresql.stop()
        


"""
Module for running taeg command-line utilities
Added because Windows can't shebang correctly and thus needs taeg to be run via `python3 -m`
"""

import sys

from textwrap import dedent

from .argparser import get_parser
from .version import print_version_info

from . import assign
from . import check
from . import export
from . import generate
from . import grade

MISSING_PACKAGES = False
try:
    from . import service
except ImportError:
    MISSING_PACKAGES = True

PARSER = get_parser()

def run_taeg(unparsed_args=None):
    args = PARSER.parse_args(unparsed_args)

    if args.version:
        print_version_info(logo=True)
        return

    if hasattr(args, 'func_str'):
        if args.func_str.startswith("service") and MISSING_PACKAGES:
            raise ImportError(
                "Missing some packages required for taeg service. "
                "Please install all requirements at "
                "https://raw.githubusercontent.com/taecoding/tae-grader/master/requirements.txt"
            )

        args.func = eval(args.func_str)
        args.func(args)

    elif len(sys.argv) > 1 and sys.argv[1] == "generate":
        print(dedent("""\
        You must specify a command for Taeg Generate:
            autograder
            token"""))

    elif len(sys.argv) > 1 and sys.argv[1] == "service":
        print(dedent("""\
        You must specify a command for Taeg Service:
            build
            create
            start"""))

    else:
        print(dedent("""\
        You must specify a command for Taeg:
            assign
            check
            export
            generate
            grade
            service"""))



"""
TAE-Grader version and printable logo
"""

import sys

from textwrap import dedent

__version__ = "1.0.1"

LOGO_WITH_VERSION = fr"""
________________  ___________   ________ 
\__    ___/  _  \ \_   _____/  /  _____/ 
  |    | /  /_\  \ |    __)_  /   \  ___ 
  |    |/    |    \|        \ \    \_\  \
  |____|\____|__  /_______  /  \______  /
                \/        \/          \/ 
                                                v{__version__}
"""

def print_version_info(logo=False):
    if logo:
        print(LOGO_WITH_VERSION + "\n")
    print(dedent(f"""\
        Python version: {".".join(str(v) for v in sys.version_info[:3])}
        TAE-Grader version: {__version__}
    """))



"""
Miscellaneous cell generators for Taeg Assign
"""

import copy
import nbformat

from .constants import MD_RESPONSE_CELL_SOURCE
from .utils import get_source, lock

def gen_init_cell():
    """
    Generates a cell to initialize Taeg in the notebook. The code cell has the following contents:

    .. code-block:: python

        # Initialize Taeg
        import taeg
        grader = taeg.Notebook()
    
    Returns:
        ``nbformat.NotebookNode``: the init cell
    """
    cell = nbformat.v4.new_code_cell("# Initialize Taeg\nimport taeg\ngrader = taeg.Notebook()")
    lock(cell)
    return cell

def gen_markdown_response_cell():
    """
    Generates a Markdown response cell with the following contents:

    .. code-block:: markdown

        _Type your answer here, replacing this text._

    Returns:
        ``nbformat.NotebookNode``: the response cell
    """
    return nbformat.v4.new_markdown_cell(MD_RESPONSE_CELL_SOURCE)

def gen_export_cells(instruction_text, assignment, pdf=True, filtering=True):
    """
    Generates export cells that instruct the student the run a code cell calling 
    ``taeg.Notebook.export`` to generate and download their submission. The Markdown cell contains:

    .. code-block:: markdown

        ## Submission
        
        Make sure you have run all cells in your notebook in order before running the cell below, so 
        that all images/graphs appear in the output. The cell below will generate a zipfile for you 
        to submit. **Please save before exporting!**

    Additional instructions can be appended to this cell by passing a string to ``instruction_text``.

    The code cell contains:

    .. code-block:: python

        # Save your notebook first, then run this cell to export your submission.
        grader.export()
    
    The call to ``grader.export()`` contains different arguments based on the values passed to ``pdf``
    and ``filtering``. 
    
    Args:
        instruction_text (``str``): extra instructions for students when exporting
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
        pdf (``bool``, optional): whether a PDF is needed
        filtering (``bool``, optional): whether PDF filtering is needed
    
    Returns:
        ``list`` of ``nbformat.NotebookNode``: generated export cells
    """
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "## Submission\n\nMake sure you have run all cells in your notebook in order before " \
    "running the cell below, so that all images/graphs appear in the output. The cell below will generate " \
    "a zip file for you to submit. **Please save before exporting!**"
    
    if instruction_text:
        instructions.source += '\n\n' + instruction_text

    export = nbformat.v4.new_code_cell()
    source_lines = ["# Save your notebook first, then run this cell to export your submission."]
    if filtering and pdf:
        source_lines.append(f"grader.export(\"{assignment.master}\")")
    elif not filtering:
        source_lines.append(f"grader.export(\"{assignment.master}\", filtering=False)")
    else:
        source_lines.append(f"grader.export(\"{assignment.master}\", pdf=False)")
    export.source = "\n".join(source_lines)

    lock(instructions)
    lock(export)

    return [instructions, export, nbformat.v4.new_markdown_cell(" ")]     # last cell is buffer

def gen_check_all_cell():
    """
    Generates a check-all cell and a Markdown cell with instructions to run all tests in the notebook. 
    The Markdown cell has the following contents:

    .. code-block:: markdown

        ---
        
        To double-check your work, the cell below will rerun all of the autograder tests.

    The code cell has the following contents:

    .. code-block:: python

        grader.check_all()
    
    Returns:
        ``list`` of ``nbformat.NotebookNode``: generated check-all cells
    """
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "---\n\nTo double-check your work, the cell below will rerun all of the autograder tests."

    check_all = nbformat.v4.new_code_cell("grader.check_all()")

    lock(instructions)
    lock(check_all)

    return [instructions, check_all]

def gen_close_export_cell():
    """
    Generates a Markdown cell to end question export for PDF filtering. The cell contains:

    .. code-block:: markdown

        <!-- END QUESTION -->
    
    Returns:
        ``nbformat.NotebookNode``: new Markdown cell with ``<!-- END QUESTION -->``
    """
    cell = nbformat.v4.new_markdown_cell("<!-- END QUESTION -->")
    lock(cell)
    return cell

def add_close_export_to_cell(cell):
    """Adds an HTML comment to close question export for PDF filtering to the top of ``cell``. ``cell``
    should be a Markdown cell. This adds ``<!-- END QUESTION-->`` as the first line of the cell.
    
    Args:
        cell (``nbformat.NotebookNode``): the cell to add the close export to

    Returns:
        ``nbformat.NotebookNode``: the cell with the close export comment at the top
    """
    cell = copy.deepcopy(cell)
    source = get_source(cell)
    source = ["<!-- END QUESTION -->\n", "\n"] + source
    cell['source'] = "\n".join(source)
    return cell




#################################
##### Tests for taeg grade #####
#################################

import os
import unittest
import subprocess
import json
import re
import pandas as pd

from unittest import mock
from subprocess import PIPE
from glob import glob

from taeg.argparser import get_parser
from taeg.grade import main as grade
from taeg.grade.metadata import GradescopeParser, CanvasParser, JSONParser, YAMLParser

from . import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test-grade/"

class TestGrade(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        create_image_cmd = ["make", "docker-test"]
        subprocess.run(create_image_cmd, check=True)
        # create_image = subprocess.run(create_image_cmd, check=True)
        # assert not create_image.stderr, create_image.stderr.decode("utf-8")
    
    def setUp(self):
        """
        Load in point values
        """
        self.test_points = {}
        for test_file in glob(TEST_FILES_PATH + "tests/*.py"):
            env = {}
            with open(test_file) as f:
                exec(f.read(), env)
            self.test_points[env['test']['name']] = env['test']['points']
        return super().setUp()


    def test_docker(self):
        """
        Check that we have the right container installed and that docker is running
        """
        # use docker image inspect to see that the image is installed and tagged as tae-grader
        inspect = subprocess.run(["docker", "image", "inspect", "taeg-test"], stdout=PIPE, stderr=PIPE)

        # assert that it didn't fail, it will fail if it is not installed
        self.assertEqual(len(inspect.stderr), 0, inspect.stderr.decode("utf-8"))


    def test_metadata_parsers(self):
        """
        Check that metadata parsers work correctly
        """
        correct_metadata = [
            {
                "identifier": "12345",
                "filename": "12345_empty_file.ipynb"
            }, {
                "identifier": "23456",
                "filename": "23456_empty_file.ipynb"
            }, {
                "identifier": "34567",
                "filename": "34567_empty_file.ipynb"
            }, {
                "identifier": "45678",
                "filename": "45678_empty_file.ipynb"
            }, {
                "identifier": "56789",
                "filename": "56789_empty_file.ipynb"
            }
        ]

        correct_file_to_id = {
            "12345_empty_file.ipynb": "12345",
            "23456_empty_file.ipynb": "23456",
            "34567_empty_file.ipynb": "34567",
            "45678_empty_file.ipynb": "45678",
            "56789_empty_file.ipynb": "56789",
        }

        correct_id_to_file = {
            "12345": "12345_empty_file.ipynb",
            "23456": "23456_empty_file.ipynb",
            "34567": "34567_empty_file.ipynb",
            "45678": "45678_empty_file.ipynb",
            "56789": "56789_empty_file.ipynb",
        }

        correct_filenames = [
            "12345_empty_file.ipynb",
            "23456_empty_file.ipynb",
            "34567_empty_file.ipynb",
            "45678_empty_file.ipynb",
            "56789_empty_file.ipynb",
        ]

        correct_identifiers = [
            "12345",
            "23456",
            "34567",
            "45678",
            "56789",
        ]

        try:
            # gradescope parser
            gs_parser = GradescopeParser(TEST_FILES_PATH + "gradescope-export")
            self.assertCountEqual(gs_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(gs_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(gs_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], gs_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], gs_parser.id_to_file(identifier))

            # canvas parser
            canvas_parser = CanvasParser(TEST_FILES_PATH + "canvas-export")
            self.assertCountEqual(canvas_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(canvas_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(canvas_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], canvas_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], canvas_parser.id_to_file(identifier))

            # JSON parser
            json_parser = JSONParser(TEST_FILES_PATH + "meta.json")
            self.assertCountEqual(json_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(json_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(json_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], json_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], json_parser.id_to_file(identifier))

            # YAML parser
            yaml_parser = YAMLParser(TEST_FILES_PATH + "meta.yml")
            self.assertCountEqual(yaml_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(yaml_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(yaml_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], yaml_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], yaml_parser.id_to_file(identifier))

            # cleanup
            gs_rm = subprocess.run(["rm", "-rf"] + glob(TEST_FILES_PATH + "gradescope-export/*.ipynb"), stdout=PIPE, stderr=PIPE)
            self.assertEqual(len(gs_rm.stderr), 0, gs_rm.stderr.decode("utf-8"))

        except:
            # cleanup
            gs_rm = subprocess.run(["rm", "-rf"] + glob(TEST_FILES_PATH + "gradescope-export/*.ipynb"), stdout=PIPE, stderr=PIPE)
            self.assertEqual(len(gs_rm.stderr), 0, gs_rm.stderr.decode("utf-8"))
            raise


    def test_notebooks(self):
        """
        Check that the example of 100 notebooks runs correctely locally.
        """
        # grade the 100 notebooks
        grade_command = ["grade",
            # NO METADATA PASSED, test case when no metadata provided
            # "-y", TEST_FILES_PATH + "notebooks/meta.yml", 
            "-p", TEST_FILES_PATH + "notebooks/", 
            "-t", TEST_FILES_PATH + "tests/", 
            "-r", TEST_FILES_PATH + "requirements.txt",
            "-o", "test/",
            "--image", "taeg-test",
            "-v"
        ]
        args = parser.parse_args(grade_command)
        args.func = grade
        args.func(args)

        # read the output and expected output
        df_test = pd.read_csv("test/final_grades.csv")
        self.assertTrue("identifier" not in df_test.columns, "did not drop identifier column when no metadata passed")

        # sort by filename
        df_test = df_test.sort_values("file").reset_index(drop=True)
        df_test["failures"] = df_test["file"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

        # check point values
        for _, row in df_test.iterrows():
            for test in self.test_points:
                if int(re.sub(r"\D", "", test)) in row["failures"]:
                    self.assertEqual(row[test], 0, "{} supposed to fail {} but passed".format(row["file"], test))
                else:
                    self.assertEqual(row[test], self.test_points[test], "{} supposed to pass {} but failed".format(row["file"], test))

        # df_correct = pd.read_csv(TEST_FILES_PATH + "final_grades_correct_notebooks.csv").sort_values("identifier").reset_index(drop=True)

        # # assert the dataframes are as expected
        # self.assertTrue(df_test.equals(df_correct), "Dataframes not equal")

        # remove the extra output
        cleanup_command = ["rm", "test/final_grades.csv"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)

        # assert cleanup worked
        self.assertEqual(len(cleanup.stderr), 0, "Error in cleanup")


    def test_notebooks_with_pdfs(self):
        """
        Check that the example of 100 notebooks runs correctely locally.
        """
        # grade the 100 notebooks
        grade_command = ["grade",
            "-y", TEST_FILES_PATH + "notebooks/meta.yml", 
            "-p", TEST_FILES_PATH + "notebooks/", 
            "-t", TEST_FILES_PATH + "tests/", 
            "-r", TEST_FILES_PATH + "requirements.txt",
            "-o", "test/",
            "--pdfs",
            "--containers", "5",
            "--image", "taeg-test"
        ]
        args = parser.parse_args(grade_command)
        args.func = grade
        args.func(args)

        # check that we have PDFs
        self.assertTrue(os.path.isdir("test/submission_pdfs"))
        for file in glob(TEST_FILES_PATH + "notebooks/*.ipynb"):
            pdf = "test/submission_pdfs/" + os.path.split(file)[1][:-5] + "pdf"
            self.assertTrue(os.path.isfile(pdf))

        # remove the extra output
        cleanup_command = ["rm", "-rf", "test/final_grades.csv", "test/submission_pdfs"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(cleanup.stderr), 0, cleanup.stderr.decode("utf-8"))


    def test_scripts(self):
        """
        Check that the example of 100 scripts runs correctely locally.
        """
        grade_command = ["grade",
            "-sy", TEST_FILES_PATH + "scripts/meta.yml", 
            "-p", TEST_FILES_PATH + "scripts/", 
            "-t", TEST_FILES_PATH + "tests/", 
            "-r", TEST_FILES_PATH + "requirements.txt",
            "-o", "test/",
            "--image", "taeg-test"
        ]
        args = parser.parse_args(grade_command)
        args.func = grade
        args.func(args)

        # read the output and expected output
        df_test = pd.read_csv("test/final_grades.csv").sort_values("identifier").reset_index(drop=True)
        df_test["failures"] = df_test["identifier"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

        # check point values
        for _, row in df_test.iterrows():
            for test in self.test_points:
                if int(re.sub(r"\D", "", test)) in row["failures"]:
                    self.assertEqual(row[test], 0, "{} supposed to fail {} but passed".format(row["identifier"], test))
                else:
                    self.assertEqual(row[test], self.test_points[test], "{} supposed to pass {} but failed".format(row["identifier"], test))
        
        # grade the 100 scripts
        
        # args = parser.parse_args(grade_command)
        # args.func = grade
        # args.func(args)

        # # read the output and expected output
        # df_test = pd.read_csv("test/final_grades.csv").sort_values("identifier").reset_index(drop=True)
        # df_correct = pd.read_csv(TEST_FILES_PATH + "final_grades_correct_script.csv").sort_values("identifier").reset_index(drop=True)

        # # assert the dataframes are as expected
        # self.assertTrue(df_test.equals(df_correct), "Dataframes not equal")

        # remove the extra output
        cleanup_command = ["rm", "test/final_grades.csv"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)

        # assert cleanup worked
        self.assertEqual(len(cleanup.stderr), 0, "Error in cleanup")



"""
Utilities for Taeg Service
"""

from psycopg2 import connect, extensions

def connect_db(host="localhost", username="admin", password="", port="5432", db="taeg_db"):
    """Connects to a specific Postgres database with provided parameters/credentials

    Arguments:
        host (``str``, optional): hostname for database (default 'localhost')
        username (``str``, optional): username with proper read/write permissions for Postgres
        password (``str``, optional): password for provided username
        port (``str``, optional): port on which Postgres is running

    Returns:
        ``psycopg2.connection``: connection object for executing SQL commands on Postgres database

    Raises:
        ``ImportError``: if psycopg2 is not installed
    """
    
    conn = connect(dbname=db,
               user=username,
               host=host,
               password=password,
               port=port)
    conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    return conn



"""
Runs Taeg on Gradescope with the configurations specified below
"""

import os
import subprocess

from taeg.generate.run_autograder import main as run_autograder

config = {
    "score_threshold": {{ threshold }},
    "points_possible": {{ points }},
    "show_stdout_on_release": {{ show_stdout }},
    "show_hidden_tests_on_release": {{ show_hidden }},
    "seed": {{ seed }},
    "grade_from_log": {{ grade_from_log }},
    "serialized_variables": {{ serialized_variables }},
    "public_multiplier": {{ public_multiplier }},
    "token": {% if token %}'{{ token }}'{% else %}None{% endif %},
    "course_id": '{{ course_id }}',
    "assignment_id": '{{ assignment_id }}',
    "filtering": {{ filtering }},
    "pagebreaks": {{ pagebreaks }},
    "debug": False,
    "autograder_dir": '{{ autograder_dir }}',
    "lang": '{{ lang }}',
}

if __name__ == "__main__":
    run_autograder(config)



####################################
##### Tests for taeg execute #####
####################################
import os
import unittest
import subprocess
import json
import shutil

from .. import TestCase
from ast import literal_eval
from taeg.execute.results import GradingResults
from taeg.test_files.abstract_test import TestCollectionResults, TestFile
from taeg.test_files.ok_test import OKTestFile

class TestResults(TestCase):

    def testGetScore(self):

        """
        Tests happy path for get_score
        """

        tests = [OKTestFile("test1", ["t1","t2"],[True,False]), OKTestFile("test2", ["t1","t2"],[True,False],value=0.9,hidden=True)]
        testRes = TestCollectionResults(grade=0.9,paths="test/",tests=tests,passed_tests=tests,failed_tests=[],include_grade=True)
        g = GradingResults([testRes])
        self.assertEqual(g.get_score("test1"), 0.9)

    def testRepr(self):

        """
        Tests repr styling of results
        """

        tests = [OKTestFile("test1", ["t1","t2"],[True,False]), OKTestFile("test2", ["t1","t2"],[True,False],value=0.9,hidden=True)]
        testRes = TestCollectionResults(grade=0.9,paths="test/",tests=tests,passed_tests=tests,failed_tests=[],include_grade=True)
        g = GradingResults([testRes])
        str = "{ 'test1': { 'hidden': False,\n             'incorrect': False,\n             'name': 'test1',\n             'possible': 1,\n             'score': 0.9,\n             'test':    test1\n\nTest result:\nNone},\n  'test2': { 'hidden': False,\n             'incorrect': False,\n             'name': 'test2',\n             'possible': 0.9,\n             'score': 0.81,\n             'test':    test2\n\nTest result:\nNone}}"
        self.assertEqual(repr(g),str)

    def testDict(self):

        """
        Tests dictionary representation of results
        """

        tests = [OKTestFile("test1", ["t1","t2"],[True,False]), OKTestFile("test2", ["t1","t2"],[True,False],value=0.9,hidden=False)]
        testRes = TestCollectionResults(grade=0.9,paths="test/",tests=tests,passed_tests=tests,failed_tests=[],include_grade=True)
        g = GradingResults([testRes])
        d = "{'test1': {'name': 'test1', 'score': 0.9, 'possible': 1, 'test':    test1\n\nTest result:\nNone, 'hidden': False, 'incorrect': False}, 'test2': {'name': 'test2', 'score': 0.81, 'possible': 0.9, 'test':    test2\n\nTest result:\nNone, 'hidden': False, 'incorrect': False}}"
        self.assertEqual(str(g.to_dict()),d)

    def testPublicScore(self):

        """
        Tests public scoring of results
        """

        tests = [OKTestFile("test1", ["t1","t2"],[True,False]), OKTestFile("test2", ["t1","t2"],[True,False],value=0.9,hidden=False)]
        testRes = TestCollectionResults(grade=0.9,paths="test/",tests=tests,passed_tests=tests,failed_tests=[],include_grade=True)
        g = GradingResults([testRes])
        self.assertEqual(g.get_public_score("test1"), 1)
        self.assertEqual(g.get_public_score("test2"), 0.9)

    def testGradescopeDict(self):

        """
        Tests gradescope config edge cases for gradescope dict 
        """

        tests = [OKTestFile("test1", ["t1","t2"],[True,False]), OKTestFile("test2", ["t1","t2"],[True,False],value=0.9,hidden=False)]
        testRes = TestCollectionResults(grade=0.9,paths="test/",tests=tests,passed_tests=tests,failed_tests=[],include_grade=True)
        g = GradingResults([testRes])
        # TODO : multiline this

        gs_dict = "{'tests': [{'name': 'test1 - Public', 'score': 0.0, 'max_score': 0, 'visibility': 'visible', 'output': 'All tests passed!'}, {'name': 'test1 - Hidden', 'score': 0.9, 'max_score': 1, 'visibility': 'hidden', 'output': 'All tests passed!'}, {'name': 'test2 - Public', 'score': 0.0, 'max_score': 0.0, 'visibility': 'visible', 'output': 'All tests passed!'}, {'name': 'test2 - Hidden', 'score': 0.81, 'max_score': 0.9, 'visibility': 'hidden', 'output': 'All tests passed!'}], 'stdout_visibility': 'after_published', 'score': 2}"
        self.assertEqual(str(g.to_gradescope_dict(config={"show_stdout_on_release":True,"score_threshold":0.5,"points_possible":2})), gs_dict)        



"""
Assignment configurations for Taeg Assign
"""

import yaml

from .constants import BLOCK_QUOTE
from .utils import get_source, get_spec

class Assignment:
    """
    A class that houses configurations for an assignment. Contains a dictionary of default arguments
    that can be updated in an instance using the ``update()`` method. Functions similarly to an 
    ``AttrDict`` in that keys of the configuration can be accessed as ``assignment.<key>``.

    To access a configuration value, use the dot syntax. For example, to access the ``generate`` key
    of an ``Assignment`` instance ``assignment``:
    
    .. code-block::python
        
        assignment.generate
    
    If ``generate`` is present in ``assignment.config``, then the value in that dictionary will be 
    returned. If it is not, the value in ``Assignment.defaults`` will be returned instead. Configurations
    can also be updated using dot syntax:
    
    .. code-block:: python
        
        assignment.generate = True

    If a key not present in ``Assignment.defaults`` is attempted to be accessed or set, an 
    ``AttributeError`` will be thrown.

    Attributes:
        config (``dict``): the configurations specific to this assignment; keys in this dictionary
            are used before the defaults if present.
    """
    defaults = {
        "master": None,
        "result": None,
        "generate": {},
        "solutions_pdf": False,
        "template_pdf": False,
        "seed_required": False,
        "service": {},
        "save_environment": False,
        "requirements": None,
        "overwrite_requirements": False,
        "files": [],
        "variables": {},
        "run_tests": True,
        "ignore_modules": [],
        "init_cell": True,
        "check_all_cell": True,
        "export_cell": True,
        "seed": None,
        "lang": None,
    }

    def __init__(self):
        self.config = type(self).defaults.copy()

    def __getattr__(self, attr):
        if attr in type(self).defaults:
            return self.config.get(attr, type(self).defaults[attr])
        raise AttributeError(f"Assignment has no attribute {attr}")

    def __setattr__(self, attr, value):
        if attr == "config":
            self.__dict__[attr] = value
        elif attr in type(self).defaults:
            self.config[attr] = value
        else:
            raise AttributeError(f"Assignment has no attribute {attr}")

    def update(self, config):
        """
        Updates the configuration stored in this assignment using keys and values in the dictionary
        ``config``

        Args:
            config (``dict``): new configurations
        """
        self.config.update(config)

    @property
    def is_r(self):
        return self.lang == "r"
    
    @property
    def is_python(self):
        return self.lang == "python"

    @property
    def is_rmd(self):
        return self.master.suffix.lower() == ".rmd"

def read_assignment_metadata(cell):
    """
    Return assignment metadata from an assignment cell
    
    Args:
        cell (``nbformat.NotebookNode``): the assignment cell
    
    Returns:
        ``dict``: assignment metadata
    """
    source = get_source(cell)
    begin_assignment_line = get_spec(source, "assignment")
    i, lines = begin_assignment_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    metadata = yaml.full_load('\n'.join(lines))
    return metadata

def is_assignment_cell(cell):
    """
    Returns whether cell contains BEGIN ASSIGNMENT in a block quote
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``bool``: whether the current cell is an assignment definition cell
    """
    if cell.cell_type != 'markdown':
        return False
    return get_spec(get_source(cell), "assignment") is not None



"""
Gradescope autograding internals for Python
"""

import os
import json
import shutil
import subprocess
import warnings

from glob import glob

from ..constants import DEFAULT_OPTIONS
from ..token import APIClient
from ..utils import replace_notebook_instances

from ...check.logs import Log, QuestionNotInLogException
from ...check.notebook import _TAEG_LOG_FILENAME
from ...execute import grade_notebook
from ...export import export_notebook

def prepare_files():
    # put files into submission directory
    if os.path.exists("./source/files"):
        for file in os.listdir("./source/files"):
            fp = os.path.join("./source/files", file)
            if os.path.isdir(fp):
                if not os.path.exists(os.path.join("./submission", os.path.basename(fp))):
                    shutil.copytree(fp, os.path.join("./submission", os.path.basename(fp)))
            else:
                shutil.copy(fp, "./submission")

    # create __init__.py files
    subprocess.run(["touch", "./__init__.py"])
    subprocess.run(["touch", "./submission/__init__.py"])

    os.makedirs("./submission/tests", exist_ok=True)
    tests_glob = glob("./source/tests/*.py")
    for file in tests_glob:
        shutil.copy(file, "./submission/tests")

def write_and_submit_pdf(client, nb_path, filtering, pagebreaks, course_id, assignment_id):
    try:
        export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
        pdf_path = os.path.splitext(nb_path)[0] + ".pdf"

        # get student email
        with open("../submission_metadata.json") as f:
            metadata = json.load(f)

        student_emails = []
        for user in metadata["users"]:
            student_emails.append(user["email"])
        
        for student_email in student_emails:
            client.upload_pdf_submission(course_id, assignment_id, student_email, pdf_path)

        print("\n\nSuccessfully uploaded submissions for: {}".format(", ".join(student_emails)))

    except Exception as e:
        # print("\n\n")
        # warnings.warn("PDF generation or submission failed", RuntimeWarning)
        print(f"\n\nError encountered while generating and submitting PDF:\n{e}")

def run_autograder(config):
    """
    Runs autograder on Gradescope based on predefined configurations.

    Args:
        config (``dict``): configurations for autograder
    """
    options = DEFAULT_OPTIONS.copy()
    options.update(config)

    # add miniconda back to path
    os.environ["PATH"] = f"{options['miniconda_path']}/bin:" + os.environ.get("PATH")
    
    abs_ag_path = os.path.abspath(options["autograder_dir"])
    os.chdir(abs_ag_path)

    if options["token"] is not None:
        client = APIClient(token=options["token"])
        generate_pdf = True
    else:
        generate_pdf = False

    prepare_files()

    os.chdir("./submission")

    nb_path = glob("*.ipynb")[0]

    replace_notebook_instances(nb_path)

    if os.path.isfile(_TAEG_LOG_FILENAME):
        log = Log.from_file(_TAEG_LOG_FILENAME, ascending=False)
        if options["grade_from_log"]:
            print("\n\n")     # for logging in taeg.execute.execute_log
    else:
        assert not options["grade_from_log"], "missing log"
        log = None

    scores = grade_notebook(
        nb_path, 
        glob("./tests/*.py"), 
        name="submission", 
        cwd=".", 
        test_dir="./tests",
        ignore_errors=not options["debug"], 
        seed=options["seed"],
        log=log if options["grade_from_log"] else None,
        variables=options["serialized_variables"]
    )

    # verify the scores against the log
    print("\n\n")
    if log is not None:
        try:
            found_discrepancy = scores.verify_against_log(log)
            if not found_discrepancy:
                print("No discrepancies found while verifying scores against the log.")
        except BaseException as e:
            print(f"Error encountered while trying to verify scores with log:\n{e}")
    else:
        print("No log found with which to verify student scores")

    if generate_pdf:
        write_and_submit_pdf(
            client, nb_path, options['filtering'], options['pagebreaks'], options['course_id'], 
            options['assignment_id']
        )

    output = scores.to_gradescope_dict(config)

    os.chdir(abs_ag_path)

    return output



"""
Master notebook parser and transformer for Taeg Assign
"""

import os
import copy
import pathlib
import nbformat

from .assignment import is_assignment_cell, read_assignment_metadata
from .cell_generators import (
    gen_init_cell, gen_markdown_response_cell, gen_export_cells, gen_check_all_cell, 
    gen_close_export_cell, add_close_export_to_cell
)
from .questions import is_question_cell, read_question_metadata, gen_question_cell
from .solutions import is_markdown_solution_cell, has_seed
from .tests import is_test_cell, any_public_tests
from .utils import is_markdown_cell, EmptyCellException

def transform_notebook(nb, assignment, args):
    """
    Converts a master notebook to an Taeg-formatted solutions notebook, parsing test cells into
    dictionaries ready to be written as OK test files.

    Args:
        nb (``nbformat.NotebookNode``): the master notebook
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
        args (``argparse.Namespace``): parsed command line arguments

    Returns:
        ``tuple(nbformat.NotebookNode, dict)``: the transformed notebook and a dictionary mapping 
            test names to their parsed contents

    """
    transformed_cells, test_files = get_transformed_cells(nb['cells'], assignment)

    if assignment.init_cell and not args.no_init_cell and assignment.is_python:
        transformed_cells = [gen_init_cell()] + transformed_cells

    if assignment.check_all_cell and not args.no_check_all and assignment.is_python:
        transformed_cells += gen_check_all_cell()
    
    if assignment.export_cell and not args.no_export_cell and assignment.is_python:
        export_cell = assignment.export_cell
        if export_cell is True:
            export_cell = {}

        transformed_cells += gen_export_cells(
            export_cell.get('instructions', ''), 
            assignment,
            pdf = export_cell.get('pdf', True),
            filtering = export_cell.get('filtering', True)
        )

    transformed_nb = copy.deepcopy(nb)
    transformed_nb['cells'] = transformed_cells

    return transformed_nb, test_files

def get_transformed_cells(cells, assignment):
    """
    Takes in a list of cells from the master notebook and returns a list of cells for the solutions
    notebook. Replaces test cells with a cell calling ``taeg.Notebook.check``, inserts Markdown
    response cells for manual questions with Markdown solutions, and comments out question metadata 
    in question cells, among other things.

    Args:
        cells (``list`` of ``nbformat.NotebookNode``): original code cells
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
    
    Returns:
        ``tuple(list, dict)``: list of cleaned notebook cells and a dictionary mapping test names to 
            their parsed contents
    """
    if assignment.is_r:
        from .r_adapter.tests import read_test, gen_test_cell
    else:
        from .tests import read_test, gen_test_cell
    
    # global SEED_REQUIRED, ASSIGNMENT_METADATA
    transformed_cells, test_files = [], {}
    question_metadata, test_cases, processed_solution, md_has_prompt = {}, [], False, False
    need_close_export, no_solution = False, False

    for cell in cells:

        if has_seed(cell):
            assignment.seed_required = True            

        # this is the prompt cell or if a manual question then the solution cell
        if question_metadata and not processed_solution:
            assert not is_question_cell(cell), f"Found question cell before end of previous question cell: {cell}"

            # if this isn't a MD solution cell but in a manual question, it has a Markdown prompt
            if question_metadata.get('manual', False) and is_markdown_cell(cell) and not is_markdown_solution_cell(cell):
                md_has_prompt = True
                transformed_cells.append(cell)
                continue
            
            # if this a manual question but not MD solution, it has a code solution cell
            elif question_metadata.get('manual', False) and not is_markdown_solution_cell(cell):
                md_has_prompt = True
            
            # if there is no prompt, add a prompt cell
            elif is_markdown_solution_cell(cell) and not md_has_prompt:
                transformed_cells.append(gen_markdown_response_cell())

            # if this is a test cell, this question has no response cell for the students, so we don't
            # include it in the output notebook but we need a test file
            elif is_test_cell(cell):
                no_solution = True
                test = read_test(cell, question_metadata, assignment)
                test_cases.append(test)

            # elif is_seed_cell(cell):
            #     assignment.seed_required = True
            #     continue

            if not no_solution:
                transformed_cells.append(cell)
            
            processed_solution = True

        # if this is a test cell, parse and add to test_cases
        elif question_metadata and processed_solution and is_test_cell(cell):
            test = read_test(cell, question_metadata, assignment)
            test_cases.append(test)

        # # if this is a solution cell, append. if manual question and no prompt, also append prompt cell
        # elif question_metadata and processed_solution and is_solution_cell(cell):
        #     if is_markdown_solution_cell(cell) and not md_has_prompt:
        #         transformed_cells.append(gen_markdown_response_cell())
        #     transformed_cells.append(cell)

        else:
            # the question is over -- we've seen the question and solution and any tests and now we 
            # need to get ready to process the next question which *could be this cell*
            if question_metadata and processed_solution:

                # create a Notebook.check cell
                if test_cases:
                    check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)

                    # only add to notebook if there's a response cell or if there are public tests
                    if not no_solution or any_public_tests(test_cases):
                        transformed_cells.append(check_cell)

                # add a cell with <!-- END QUESTION --> if a manually graded question
                manual = question_metadata.get('manual', False)
                if manual:
                    need_close_export = True
                
                # reset vars
                question_metadata, processed_solution, test_cases, md_has_prompt, no_solution = {}, False, [], False, False

            # update assignment config if present; don't add cell to output nb
            if is_assignment_cell(cell):
                assignment.update(read_assignment_metadata(cell))

            # if a question cell, parse metadata, comment out question metadata, and append to nb
            elif is_question_cell(cell):
                question_metadata = read_question_metadata(cell)
                manual = question_metadata.get('manual', False)

                try:
                    transformed_cells.append(gen_question_cell(cell, manual, need_close_export))
                    need_close_export = False
                except EmptyCellException:
                    pass

            elif is_markdown_solution_cell(cell):
                transformed_cells.append(gen_markdown_response_cell())
                transformed_cells.append(cell)

            else:
                assert not is_test_cell(cell), f"Test outside of a question: {cell}"

                if need_close_export:
                    if cell['cell_type'] == 'code':
                        transformed_cells.append(gen_close_export_cell())
                    else:
                        cell = add_close_export_to_cell(cell)
                    need_close_export = False

                transformed_cells.append(cell)

    if test_cases:
        check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)
        if not no_solution or any_public_tests(test_cases):
            transformed_cells.append(check_cell)

    return transformed_cells, test_files



"""
PDF via HTML exporter
"""

import os
import tempfile
import shutil
import pdfkit
import nbconvert
import pkg_resources

from io import StringIO, BytesIO
from contextlib import redirect_stdout, redirect_stderr
from nbconvert.exporters import export
from PyPDF2 import PdfFileMerger

from .base_exporter import BaseExporter
from .utils import notebook_pdf_generator

class PDFViaHTMLExporter(BaseExporter):
    """
    Exports notebooks to PDF files using HTML as an intermediary

    Converts IPython notebooks to PDFs by first converting them into temporary HTML files that are then
    converted to PDFs using wkhtmltopdf and its Python API pdfkit which are then stitched together (if
    pagebreaks are enabled) using PyPDF2.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update({
        "save_html": False,
        "template": "templates/via_html.tpl"
    })

    @classmethod
    def convert_notebook(cls, nb_path, dest, debug=False, **kwargs):
        assert shutil.which("wkhtmltopdf") is not None, "Cannot export via HTML without wkhtmltopdf"

        options = cls.default_options.copy()
        options.update(kwargs)

        nb = cls.load_notebook(nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"])

        exporter = nbconvert.HTMLExporter()
        exporter.template_file = pkg_resources.resource_filename(__name__, options["template"])

        if options["save_html"]:
            html, _ = export(exporter, nb)
            html_path = os.path.splitext(dest)[0] + ".html"
            with open(html_path, "wb+") as f:
                f.write(html.encode("utf-8"))
        
        merger = PdfFileMerger()
        for subnb in notebook_pdf_generator(nb):
            html, _ = export(exporter, subnb)

            pdfkit_options = {
                'enable-local-file-access': None, 
                'quiet': '', 
                'print-media-type': '', 
                'javascript-delay': 2000
            }
            pdf_contents = pdfkit.from_string(html, False, options=pdfkit_options)

            output = BytesIO()
            output.write(pdf_contents)
            output.seek(0)

            merger.append(output, import_bookmarks=False)

        merger.write(dest)



##################################
##### Tests for taeg create #####
##################################

import unittest
import os
import shutil
import datetime
import testing.postgresql

from unittest import mock
from psycopg2 import connect, extensions
from psycopg2.errors import DuplicateTable

from taeg.argparser import get_parser
from taeg.service.create import create_users, remove_users
from taeg.service.create import main as create

from .. import TestCase

TEST_FILES_PATH = "test/test_service/test-create/"

parser = get_parser()

class TestCreate(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        cls.postgresql = testing.postgresql.Postgresql()
        cls.conn = connect(**cls.postgresql.dsn())
        cls.conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)

        args = parser.parse_args(["service", "create"])
        args.func = create
        args.func(args, conn=cls.conn)

        cls.conn = connect(**cls.postgresql.dsn())
        cls.conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        cls.cursor = cls.conn.cursor()
    

    def test_tables(self):
        # users table
        self.cursor.execute("""INSERT INTO users (user_id, api_keys, username, password)
                                VALUES (1234, '{{dummykey}}', 'dummyuser', 'dummypass')""")
        self.cursor.execute("""SELECT * FROM users 
                                WHERE username = 'dummyuser' """)
        output = self.cursor.fetchall()
        assert output == [(1234, [['dummykey']], 'dummyuser', 'dummypass', None)], \
            "Expected {} but was {}".format([(1234, [['dummykey']], 'dummyuser', 'dummypass', None)], output)

        # classes table
        self.cursor.execute("""INSERT INTO classes (class_id, class_name)
                                VALUES (1234, 'dummy name')""")
        self.cursor.execute("""SELECT * FROM classes 
                                WHERE class_id = '1234' """)
        assert self.cursor.rowcount == 1, "Couldn't insert/find dummy class"

        # assignments table
        self.cursor.execute("""INSERT INTO assignments (assignment_id, class_id, assignment_name)
                                VALUES ('dummyid', '1234', 'dummy name')""")
        self.cursor.execute("""SELECT * FROM assignments 
                                WHERE assignment_id = 'dummyid' """)
        output = self.cursor.fetchall()
        expected = [("dummyid", '1234', "dummy name", None)]
        assert output == expected, "Expected {} but was {}".format(expected, output)

        # submissions table
        self.cursor.execute("""INSERT INTO submissions (submission_id, assignment_id, class_id, user_id, file_path, timestamp)
                                VALUES (1234, 'dummyid', 1234, 1234, 'dummy/filepath', '{}')""".format(datetime.datetime.utcnow()))
        self.cursor.execute("""SELECT * FROM submissions 
                                WHERE submission_id = 1234 """)
        assert self.cursor.rowcount == 1, "Couldn't insert/find dummy submission"
        

    def test_user_creation(self):
        """Inserts 2 dummy username/password pairs, checks db for them,
        and removes them
        """
        user_file_path = os.path.join(TEST_FILES_PATH, "users.csv")
        create_users(user_file_path, conn=self.conn)
        # TODO: check for testusers
        self.cursor.execute("""SELECT * FROM users 
                                WHERE username = 'testuser546789' 
                                OR username = 'testuser546788'""")
        assert self.cursor.rowcount == 2, "Expected 2 rows but was {}".format(self.cursor.rowcount)
        remove_users(user_file_path, conn=self.conn)

    @classmethod
    def tearDownClass(cls):
        cls.cursor.close()
        cls.conn.close()
        cls.postgresql.stop()



"""
OK-formatted test parsers and builders for Taeg Assign
"""

import re
import pprint
import yaml
import nbformat

from collections import namedtuple

from .constants import TEST_REGEX, OTTR_TEST_NAME_REGEX, OTTR_TEST_FILE_TEMPLATE
from .utils import get_source, lock, str_to_doctest

Test = namedtuple('Test', ['input', 'output', 'hidden'])
OttrTest = namedtuple('OttrTest', ['name', 'hidden', 'body'])

def is_test_cell(cell):
    """
    Returns whether the current cell is a test cell
    
    Args:
        cell (``nbformat.NotebookNode``): a notebook cell

    Returns:
        ``bool``: whether the cell is a test cell
    """
    if cell.cell_type != 'code':
        return False
    source = get_source(cell)
    return source and re.match(TEST_REGEX, source[0], flags=re.IGNORECASE)

def any_public_tests(test_cases):
    """
    Returns whether any of the ``Test`` named tuples in ``test_cases`` are public tests.

    Args:
        test_cases (``list`` of ``Test``): list of test cases
    
    Returns:
        ``bool``: whether any of the tests are public
    """
    return any(not test.hidden for test in test_cases)

def read_test(cell, question, assignment):
    """
    Returns the contents of a test as an ``(input, output, hidden)`` named tuple
    
    Args:
        cell (``nbformat.NotebookNode``): a test cell
        question (``dict``): question metadata
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``Test`` or ``OttrTest``: test named tuple
    """
    hidden = bool(re.search("hidden", get_source(cell)[0], flags=re.IGNORECASE))
    output = ''
    for o in cell['outputs']:
        output += ''.join(o.get('text', ''))
        results = o.get('data', {}).get('text/plain')
        if results and isinstance(results, list):
            output += results[0]
        elif results:
            output += results
    return Test('\n'.join(get_source(cell)[1:]), output, hidden)

def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``taeg.Notebook.check`` on this test
    """
    cell = nbformat.v4.new_code_cell()

    cell.source = ['grader.check("{}")'.format(question['name'])]

    suites = [gen_suite(tests)]
    points = question.get('points', 1)
    
    test = {
        'name': question['name'],
        'points': points,
        'suites': suites,
    }

    tests_dict[question['name']] = test
    lock(cell)
    return cell

def gen_suite(tests):
    """
    Generates an OK test suite for a list of tests as named tuples
    
    Args:
        tests (``list`` of ``taeg.assign.Test``): test cases

    Returns:
        ``dict``: OK test suite
    """
    cases = [gen_case(test) for test in tests]
    return  {
      'cases': cases,
      'scored': True,
      'setup': '',
      'teardown': '',
      'type': 'doctest'
    }

def gen_case(test):
    """
    Generates an OK test case for a test named tuple
    
    Args:
        test (``taeg.assign.Test``): OK test for this test case

    Returns:
        ``dict``: the OK test case
    """
    code_lines = str_to_doctest(test.input.split('\n'), [])

    for i in range(len(code_lines) - 1):
        if code_lines[i+1].startswith('>>>') and len(code_lines[i].strip()) > 3 and not code_lines[i].strip().endswith("\\"):
            code_lines[i] += ';'

    code_lines.append(test.output)

    return {
        'code': '\n'.join(code_lines),
        'hidden': test.hidden,
        'locked': False
    }

def write_test(path, test):
    """
    Writes an OK test file
    
    Args:
        path (``str``): path of file to be written
        test (``dict``): OK test to be written
    """
    with open(path, 'w') as f:
        if isinstance(test, dict):
            f.write('test = ')
            pprint.pprint(test, f, indent=4, width=200, depth=None)
        else:
            f.write(test)

def remove_hidden_tests_from_dir(test_dir, assignment):
    """
    Rewrites test files in a directory to remove hidden tests
    
    Args:
        test_dir (``pathlib.Path``): path to test files directory
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
    """
    for f in test_dir.iterdir():
        if f.name == '__init__.py' or f.suffix != '.py':
            continue
        locals = {}
        with open(f) as f2:
            exec(f2.read(), globals(), locals)
        test = locals['test']
        for suite in test['suites']:
            for i, case in list(enumerate(suite['cases']))[::-1]:
                if case['hidden']:
                    suite['cases'].pop(i)
        write_test(f, test)



"""
Utilities for Taeg Assign
"""

import re
import os
import json
import pprint
import pathlib

from glob import glob

from .constants import SEED_REGEX, BLOCK_QUOTE

from ..argparser import get_parser
from ..execute import grade_notebook
from ..generate.autograder import main as generate_autograder
from ..generate.token import APIClient
from ..utils import get_relpath, get_source

class EmptyCellException(Exception):
    """
    Exception for empty cells to indicate deletion
    """


#---------------------------------------------------------------------------------------------------
# Getters
#---------------------------------------------------------------------------------------------------

def get_spec(source, begin):
    """
    Returns the line number of the spec begin line or ``None``. Converts ``begin`` to an uppercase 
    string and looks for a line matching ``f"BEGIN {begin.upper()}"``. Used for finding question and
    assignment metadata, which match ``BEGIN QUESTION`` and ``BEGIN ASSIGNMENT``, resp.
    
    Args:
        source (``list`` of ``str``): cell source as a list of lines of text
        begin (``str``): the spec to look for
    
    Returns:
        ``int``: line number of BEGIN ASSIGNMENT, if present
        ``None``: if BEGIN ASSIGNMENT not present in the cell
    """
    block_quotes = [
        i for i, line in enumerate(source) if line[:3] == BLOCK_QUOTE
    ]
    assert len(block_quotes) % 2 == 0, f"wrong number of block quote delimieters in {source}"

    begins = [
        block_quotes[i] + 1 for i in range(0, len(block_quotes), 2) 
        if source[block_quotes[i]+1].strip(' ') == f"BEGIN {begin.upper()}"
    ]
    assert len(begins) <= 1, f'multiple BEGIN {begin.upper()} blocks defined in {source}'
    
    return begins[0] if begins else None


#---------------------------------------------------------------------------------------------------
# Cell Type Checkers
#---------------------------------------------------------------------------------------------------

# def is_seed_cell(cell):
#     """
#     Returns whether ``cell`` is seed cell
    
#     Args:
#         cell (``nbformat.NotebookNode``): notebook cell
    
#     Returns:
#         ``bool``: whether the cell is a seed cell
#     """
#     if cell['cell_type'] != 'code':
#         return False
#     source = get_source(cell)
#     return source and re.match(SEED_REGEX, source[0], flags=re.IGNORECASE)

def is_markdown_cell(cell):
    """
    Returns whether ``cell`` is Markdown cell
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``bool``: whether the cell is a Markdown cell
    """
    return cell.cell_type == 'markdown'


#---------------------------------------------------------------------------------------------------
# Cell Reformatters
#---------------------------------------------------------------------------------------------------

def remove_output(nb):
    """
    Removes all outputs from a notebook in-place
    
    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb['cells']:
        if 'outputs' in cell:
            cell['outputs'] = []

def lock(cell):
    """
    Makes a cell non-editable and non-deletable in-place

    Args:
        cell (``nbformat.NotebookNode``): cell to be locked
    """
    m = cell['metadata']
    m["editable"] = False
    m["deletable"] = False


#---------------------------------------------------------------------------------------------------
# Miscellaneous
#---------------------------------------------------------------------------------------------------

def str_to_doctest(code_lines, lines):
    """
    Converts a list of lines of Python code ``code_lines`` to a list of doctest-formatted lines ``lines``

    Args:
        code_lines (``list``): list of lines of python code
        lines (``list``): set of characters used to create function name
    
    Returns:
        ``list`` of ``str``: doctest formatted list of lines
    """
    if len(code_lines) == 0:
        return lines
    line = code_lines.pop(0)
    if line.startswith(" ") or line.startswith("\t"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif line.startswith("except:") or line.startswith("elif ") or line.startswith("else:") or line.startswith("finally:"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif len(lines) > 0 and lines[-1].strip().endswith("\\"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    else:
        return str_to_doctest(code_lines, lines + [">>> " + line])

def run_tests(nb_path, debug=False, seed=None):
    """
    Runs tests in the autograder version of the notebook
    
    Args:
        nb_path (``pathlib.Path``): path to iPython notebooks
        debug (``bool``, optional): ``True`` if errors should not be ignored
        seed (``int``, optional): random seed for notebook execution
    """
    curr_dir = os.getcwd()
    os.chdir(nb_path.parent)
    # print(os.getcwd())
    results = grade_notebook(
        nb_path.name, glob(os.path.join("tests", "*.py")), cwd=os.getcwd(), 
    	test_dir=os.path.join(os.getcwd(), "tests"), ignore_errors = not debug, seed=seed
    )
    assert results.total == results.possible, "Some autograder tests failed:\n\n" + pprint.pformat(results, indent=2)
    os.chdir(curr_dir)

def write_taeg_config_file(master, result, assignment):
    """
    Creates an Taeg configuration file (a ``.taeg`` file) for students to use Taeg tools, including
    saving environments and submitting to an Taeg Service deployment, using assignment configurations.
    Writes the resulting file to the ``autograder`` and ``student`` subdirectories of ``result``.

    Args:
        master (``pathlib.Path``): path to master notebook
        result (``pathlib.Path``): path to result directory
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
    """
    config = {}

    service = assignment.service
    if service:
        config.update({
            "endpoint": service["endpoint"],
            "auth": service.get("auth", "google"),
            "assignment_id": service["assignment_id"],
            "class_id": service["class_id"]
        })

    config["notebook"] = service.get('notebook', master.name)
    config["save_environment"] = assignment.save_environment
    config["ignore_modules"] = assignment.ignore_modules

    if assignment.variables:
        config["variables"] = assignment.variables

    config_name = master.stem + '.taeg'
    with open(result / 'autograder' / config_name, "w+") as f:
        json.dump(config, f, indent=4)
    with open(result / 'student' / config_name, "w+") as f:
        json.dump(config, f, indent=4)

def run_generate_autograder(result, assignment, args):
    """
    Runs Taeg Generate on the autograder directory to generate a Gradescope zip file. Relies on 
    configurations in ``assignment.generate`` and ``args``.

    Args:
        result (``pathlib.Path``): the path to the result directory
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
        args (``argparse.Namespace``): parsed command line arguments
    """
    generate_args = assignment.generate
    if generate_args is True:
        generate_args = {}

    curr_dir = os.getcwd()
    os.chdir(str(result / 'autograder'))
    generate_cmd = ["generate", "autograder"]

    if generate_args.get('points', None) is not None:
        generate_cmd += ["--points", str(generate_args.get('points', None))]
    
    if generate_args.get('threshold', None) is not None:
        generate_cmd += ["--threshold", str(generate_args.get('threshold', None))]
    
    if generate_args.get('show_stdout', False):
        generate_cmd += ["--show-stdout"]
    
    if generate_args.get('show_hidden', False):
        generate_cmd += ["--show-hidden"]
    
    if generate_args.get('grade_from_log', False):
        generate_cmd += ["--grade-from-log"]
    
    if generate_args.get('seed', None) is not None:
        generate_cmd += ["--seed", str(generate_args.get('seed', None))]

    if generate_args.get('public_multiplier', None) is not None:
        generate_cmd += ["--public-multiplier", str(generate_args.get('public_multiplier', None))]

    if generate_args.get('pdfs', {}):
        pdf_args = generate_args.get('pdfs', {})
        token = APIClient.get_token()
        generate_cmd += ["--token", token]
        generate_cmd += ["--course-id", str(pdf_args["course_id"])]
        generate_cmd += ["--assignment-id", str(pdf_args["assignment_id"])]

        if not pdf_args.get("filtering", True):
            generate_cmd += ["--unfiltered-pdfs"]
    
    if assignment.is_r:
        generate_cmd += ["-l", "r"]

    # requirements = assignment.requirements or args.requirements
    # requirements = get_relpath(result / 'autograder', pathlib.Path(requirements))
    if assignment.requirements:
        requirements = get_relpath(result / 'autograder', pathlib.Path(assignment.requirements))
        generate_cmd += ["-r", str(requirements)]
        if assignment.overwrite_requirements or args.overwrite_requirements:
            generate_cmd += ["--overwrite-requirements"]
    
    if assignment.files or args.files:
        generate_cmd += assignment.files or args.files

    if assignment.variables:
        generate_cmd += ["--serialized-variables", str(assignment.variables)]
    
    # TODO: change this to import and direct call
    parser = get_parser()
    args = parser.parse_args(generate_cmd)
    generate_autograder(args)

    os.chdir(curr_dir)



"""
Default configurations for Gradescope autograding
"""

DEFAULT_OPTIONS = {
    "score_threshold": None,
    "points_possible": None,
    "show_stdout_on_release": False,
    "show_hidden_tests_on_release": False,
    "seed": None,
    "grade_from_log": False,
    "serialized_variables": {},
    "public_multiplier": 0,
    "token": None,
    "course_id": 'None',
    "assignment_id": 'None',
    "filtering": True,
    "pagebreaks": True,
    "debug": False,
    "autograder_dir": '/autograder',
    "lang": 'python',
    "miniconda_path": '/root/miniconda3',
    "test_visibility": 'hidden',
    "separate_tests": True,
}



"""
Execution of a Python script
"""

import os
import re
import ast

from contextlib import redirect_stdout, redirect_stderr

from .check_wrapper import CheckCallWrapper
from ..utils import hide_outputs

def execute_script(script, secret='secret', initial_env=None, ignore_errors=False, cwd=None, test_dir=None, seed=None):
    """
    Executes code of a Python script and returns the resulting global environment

    Execute script & return the global environment that results from execution. If ``ignore_errors`` is 
    ``True``, exceptions are swallowed. ``secret`` contains random digits so ``check_results`` and 
    ``check`` are not easily modifiable. ``script`` is passed in as a string.

    Args:
        script (``str``): string representation of Python script code
        secret (``str``, optional): randomly generated integer used to rebind check function
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
    
    Results:
        dict: global environment resulting from executing all code of the input script
    """
    with hide_outputs():
        if initial_env:
            global_env = initial_env.copy()
        else:
            global_env = {}
        source = ""
        # if gradescope:
        #     source = "import sys\nsys.path.append(\"/autograder/submission\")\n"
        # el
        if cwd:
            source =  f"import sys\nsys.path.append(\"{cwd}\")\n"
        if seed is not None:
            # source += "import numpy as np\nimport random\n"
            import numpy as np
            import random
            global_env["np"] = np
            global_env["random"] = random
            source += "np.random.seed({})\nrandom.seed({})\n".format(seed, seed)

        lines = script.split("\n")
        for i, line in enumerate(lines):
            # TODO: move this check into CheckCallWrapper
            if re.search(r"taeg\.Notebook\(.*?\)", line):
                # if gradescope:
                #     line = re.sub(r"taeg\.Notebook\(.*?\)", "taeg.Notebook(\"/autograder/submission/tests\")", line)
                # else:
                if test_dir:
                    line = re.sub(r"taeg\.Notebook\(.*?\)", f"taeg.Notebook(\"{test_dir}\")", line)
                else:
                    line = re.sub(r"taeg\.Notebook\(.*?\)", "taeg.Notebook(\"/home/tests\")", line)
            lines[i] = line
        try:
            exec("\n".join(lines), global_env)
            source += "\n".join(lines)
        except:
            if not ignore_errors:
                raise
        
        tree = ast.parse(source)
        # # CODE BELOW COMMENTED OUT BECAUSE the only check function is within the Notebook class
        # if find_check_assignment(tree) or find_check_definition(tree):
        #     # an empty global_env will fail all the tests
        #     return global_env

        # wrap check(..) calls into a check_results_X.append(check(..))
        transformer = CheckCallWrapper(secret)
        tree = transformer.visit(tree)
        ast.fix_missing_locations(tree)
        cleaned_source = compile(tree, filename="nb-ast", mode="exec")
        try:
            with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
                exec(cleaned_source, global_env)
        except:
            if not ignore_errors:
                raise
        return global_env




"""
Gradescope autograding internals
"""

import os
import json
import pandas as pd

from ...version import LOGO_WITH_VERSION

def main(config):
    """
    Runs autograder on Gradescope based on predefined configurations.

    Args:
        config (``dict``): configurations for autograder
    """
    print(LOGO_WITH_VERSION, "\n")

    curr_dir = os.getcwd()

    if "lang" in config and config["lang"].lower() == "r":
        from .r_adapter.run_autograder import run_autograder
    
    else:
        from .run_autograder import run_autograder
    
    output = run_autograder(config)

    with open("./results/results.json", "w+") as f:
        json.dump(output, f, indent=4)

    print("\n\n")

    df = pd.DataFrame(output["tests"])
    if "output" in df.columns:
        df.drop(columns=["output"], inplace=True)

    print(df)

    os.chdir(curr_dir)



"""
Argument parser for Taeg command-line tools
"""

import sys
import argparse

INVOKED_FROM_PYTHON = "__main__.py" in sys.argv[0]
PROG = ("taeg", "python3 -m taeg")[INVOKED_FROM_PYTHON]

# from . import assign
# from . import check
# from . import export
# from . import generate
# from . import grade
# from . import service

def get_parser():
    """Creates and returns the argument parser for Taeg
    
    Returns:
        ``argparse.ArgumentParser``: the argument parser for Taeg command-line tools
    """

    parser = argparse.ArgumentParser(prog=PROG, description="""
    A Python-based autograder for Jupyter Notebooks and Python scripts that runs locally on the instructors machine.
    Also supports use of Gradescope's autograding service, assignment distribution with taeg-assign, and public tests
    that students can run while working on assignments.
    """)
    parser.add_argument("--version", default=False, action="store_true", help="Show version information and exit")
    subparsers = parser.add_subparsers()


    ##### PARSER FOR taeg assign #####
    assign_parser = subparsers.add_parser("assign", description="Create distribution versions of taeg-assign-formatted notebook")
    assign_parser.add_argument("master", help="Notebook with solutions and tests.")
    assign_parser.add_argument("result", help="Directory containing the result.")
    assign_parser.add_argument("-l", "--lang", default=None, choices=["python", "r"], nargs="?", help="Assignment programming language; defaults to Python")
    assign_parser.add_argument("--no-export-cell", help="Don't inject an export cell into the notebook", default=False, action="store_true")
    assign_parser.add_argument("--no-run-tests", help="Don't run tests.", default=False, action="store_true")
    assign_parser.add_argument("--no-init-cell", help="Don't automatically generate an Taeg init cell", default=False, action="store_true")
    assign_parser.add_argument("--no-check-all", help="Don't automatically add a check_all cell", default=False, action="store_true")
    assign_parser.add_argument("--no-pdfs", help="Don't generate PDFs; overrides assignment config", default=False, action="store_true")
    # assign_parser.add_argument("--no-filter", help="Don't filter the PDF.", default=False, action="store_true")
    # assign_parser.add_argument("--instructions", help="Additional submission instructions for students")
    # assign_parser.add_argument("--jassign", default=False, action="store_true", help="Use jassign output notebook format")
    assign_parser.add_argument("--debug", default=False, action="store_true", help="Do not ignore errors in running tests for debugging")

    # generate options -- COMMENTED OUT BECAUSE YOU SHOULD USE THE ASSIGNMENT METADATA FOR THIS
    # assign_parser.add_argument("--generate", default=False, action="store_true", help="Generate Gradescope autograder zipfile")
    assign_parser.add_argument("-r", "--requirements", nargs='?', default=None, help="Path to requirements.txt file; ignored if no generate key in assignment metadata")
    assign_parser.add_argument("--overwrite-requirements", default=False, action="store_true", help="Overwrite (rather than append to) default requirements for Gradescope; ignored if no REQUIREMENTS argument")
    # assign_parser.add_argument("--threshold", type=float, default=None, help="Pass/fail score threshold; use with --generate only")
    # assign_parser.add_argument("--points", type=float, default=None, help="Points possible, overrides sum of test points; use with --generate only")
    # assign_parser.add_argument("--seed", type=int, default=None, help="A random seed to be executed before each cell; use with --generate only")
    # assign_parser.add_argument("--show-stdout", action="store_true", default=False, help="Show autograder test results (P/F only, no hints) after publishing grades (incl. hidden tests)")
    # assign_parser.add_argument("--show-hidden", action="store_true", default=False, help="Show autograder results for hidden tests after publishing grades")
    # assign_parser.add_argument("--grade-from-log", default=False, action="store_true", help="Whether to grade assignments based on the logged environments")

    assign_parser.add_argument("files", nargs='*', help="Other support files needed for distribution (e.g. .py files, data files)")

    assign_parser.set_defaults(func_str="assign.main")


    ##### PARSER FOR taeg check #####
    check_parser = subparsers.add_parser("check", description="Checks Python file against tests")
    check_parser.add_argument("file", help="Python file to grade")
    check_parser.add_argument("-q", "--question", help="Grade a specific test")
    check_parser.add_argument("-t", "--tests-path", default="tests", help="Path to test files")
    check_parser.add_argument("--seed", type=int, default=None, help="A random seed to be executed before each cell")

    check_parser.set_defaults(func_str="check.main")


    ##### PARSER FOR taeg export #####
    export_parser = subparsers.add_parser("export", description="Exports a Jupyter Notebook to PDF with optional filtering")
    export_parser.add_argument("source", help="Notebook to export")
    export_parser.add_argument("dest", nargs='?', default=None, help="Path to write PDF")
    export_parser.add_argument("--filtering", default=False, action="store_true", help="Whether the PDF should be filtered")
    export_parser.add_argument("--pagebreaks", default=False, action="store_true", help="Whether the PDF should have pagebreaks between questions")
    export_parser.add_argument("-s", "--save", default=False, action="store_true", help="Save intermediate file(s) as well")
    export_parser.add_argument("-e", "--exporter", default=None, choices=["latex", "html"], nargs="?", help="Type of PDF exporter to use")
    export_parser.add_argument("--debug", default=False, action="store_true", help="Export in debug mode")

    export_parser.set_defaults(func_str="export.main")


    ##### PARSER FOR taeg generate #####
    generate_parser = subparsers.add_parser("generate", description="Generates zipfile to configure Gradescope autograder")
    generate_subparsers = generate_parser.add_subparsers()


    ##### PARSER FOR taeg generate autograder #####
    generate_autograder_parser = generate_subparsers.add_parser("autograder", description="Create an autograder zip file for Gradescope")
    generate_autograder_parser.add_argument("-t", "--tests-path", nargs='?', type=str, default="./tests/", help="Path to test files")
    generate_autograder_parser.add_argument("-o", "--output-path", nargs='?', type=str, default="./", help="Path to which to write zipfile")
    generate_autograder_parser.add_argument("-r", "--requirements", nargs='?', default=None, help="Path to requirements.txt file; ./requirements.txt automatically checked")
    generate_autograder_parser.add_argument("--overwrite-requirements", default=False, action="store_true", help="Overwrite (rather than append to) default requirements for Gradescope; ignored if no REQUIREMENTS argument")
    generate_autograder_parser.add_argument("-l", "--lang", default="python", type=str, help="Assignment programming language; defaults to Python")
    generate_autograder_parser.add_argument("--threshold", type=float, default=None, help="Pass/fail score threshold")
    generate_autograder_parser.add_argument("--points", type=float, default=None, help="Points possible, overrides sum of test points")
    generate_autograder_parser.add_argument("--show-stdout", action="store_true", default=False, help="Show autograder test results (P/F only, no hints) after publishing grades (incl. hidden tests)")
    generate_autograder_parser.add_argument("--show-hidden", action="store_true", default=False, help="Show autograder results for hidden tests after publishing grades")
    generate_autograder_parser.add_argument("--seed", type=int, default=None, help="A random seed to be executed before each cell")
    generate_autograder_parser.add_argument("--token", default="", nargs="?", help="Gradescope token for generating and uploading PDFs")
    generate_autograder_parser.add_argument("--unfiltered-pdfs", default=False, action="store_true", help="Whether the PDFs should be unfiltered")
    generate_autograder_parser.add_argument("--no-pagebreaks", default=False, action="store_true", help="Whether the PDFs should not have page breaks between questions")
    generate_autograder_parser.add_argument("--course-id", default=None, help="Gradescope course ID")
    generate_autograder_parser.add_argument("--assignment-id", default=None, help="Gradescope assignment ID for PDFs")
    generate_autograder_parser.add_argument("--grade-from-log", default=False, action="store_true", help="Whether to grade assignments based on the logged environments")
    generate_autograder_parser.add_argument("--serialized-variables", default="{}", help="String representation of Python dict mapping variable names to full types for verification when deserializing log")
    generate_autograder_parser.add_argument("--public-multiplier", nargs="?", default=0, type=float, help="Percentage of points to award for passing all public tests")
    generate_autograder_parser.add_argument("--autograder-dir", nargs="?", default="/autograder", help="Root autograding directory inside grading container")
    generate_autograder_parser.add_argument("files", nargs='*', help="Other support files needed for grading (e.g. .py files, data files)")

    generate_autograder_parser.set_defaults(func_str="generate.autograder.main")


    ##### PARSER FOR taeg generate token #####
    generate_token_parser = generate_subparsers.add_parser("token", description="Get a Gradescope token")

    generate_token_parser.set_defaults(func_str="generate.token.main")


    ##### PARSER FOR taeg grade #####
    grade_parser = subparsers.add_parser("grade", description="Grade assignments locally using Docker containers")

    # necessary path arguments
    grade_parser.add_argument("-p", "--path", type=str, default="./", help="Path to directory of submissions")
    grade_parser.add_argument("-t", "--tests-path", type=str, default="./tests/", help="Path to directory of tests")
    grade_parser.add_argument("-o", "--output-path", type=str, default="./", help="Path to which to write output")

    # metadata parser arguments
    grade_parser.add_argument("-g", "--gradescope", action="store_true", default=False, help="Flag for Gradescope export")
    grade_parser.add_argument("-c", "--canvas", action="store_true", default=False, help="Flag for Canvas export")
    grade_parser.add_argument("-j", "--json", default=False, help="Flag for path to JSON metadata")
    grade_parser.add_argument("-y", "--yaml", default=False, help="Flag for path to YAML metadata")

    # submission format arguments
    grade_parser.add_argument("-s", "--scripts", action="store_true", default=False, help="Flag to incidicate grading Python scripts")
    grade_parser.add_argument("-z", "--zips", action="store_true", default=False, help="Whether submissions are zip files from Notebook.export")

    # PDF export options
    grade_parser.add_argument("--pdfs", default=False, const="unfiltered", choices=["unfiltered", "html"], nargs="?")
    # grade_parser.add_argument("--pdf", action="store_true", default=False, help="Create unfiltered PDFs for manual grading")
    # grade_parser.add_argument("--tag-filter", action="store_true", default=False, help="Create a tag-filtered PDF for manual grading")
    # grade_parser.add_argument("--html-filter", action="store_true", default=False, help="Create an HTML comment-filtered PDF for manual grading")

    # other settings and optional arguments
    grade_parser.add_argument("-f", "--files", nargs="+", help="Specify support files needed to execute code (e.g. utils, data files)")
    grade_parser.add_argument("-v", "--verbose", action="store_true", help="Flag for verbose output")
    grade_parser.add_argument("--seed", type=int, default=None, help="A random seed to be executed before each cell")
    grade_parser.add_argument("-r", "--requirements", default=None, type=str, help="Flag for Python requirements file path; ./requirements.txt automatically checked")
    grade_parser.add_argument("--containers", type=int, help="Specify number of containers to run in parallel")
    grade_parser.add_argument("--image", default="taecoding/tae-grader", help="Custom docker image to run on")
    grade_parser.add_argument("--no-kill", action="store_true", default=False, help="Do not kill containers after grading")
    grade_parser.add_argument("--debug", action="store_true", default=False, help="Print stdout/stderr from grading for debugging")

    grade_parser.set_defaults(func_str="grade.main")


    
    ###### PARSER FOR taeg service #####
    service_parser = subparsers.add_parser("service", description="Create and manage an taeg-service")
    service_subparsers = service_parser.add_subparsers()


    ##### PARSER FOR taeg service build #####
    service_build_parser = service_subparsers.add_parser("build", description="Build images for an taeg-service instance")
    service_build_parser.add_argument("repo_path", default=".", help="Path to assignments repo root")
    service_build_parser.add_argument("--db-host", default="localhost", help="Postgres database host")
    service_build_parser.add_argument("--db-port", default=5432, type=int, help="Postgres database port")
    service_build_parser.add_argument("-u", "--db-user", default="root", help="Postgres database user")
    service_build_parser.add_argument("-p", "--db-pass", default="root", help="Postgres database password")
    service_build_parser.add_argument("--image", default="taecoding/tae-grader", help="Based image for grading containers")
    service_build_parser.add_argument("-q", "--quiet", default=False, action="store_true", help="Build images without writing Docker messages to stdout")

    service_build_parser.set_defaults(func_str="service.build.main")


    ##### PARSER FOR taeg service create #####
    service_create_parser = service_subparsers.add_parser("create", description="Create database for taeg-service")
    service_create_parser.add_argument("--db-host", default="localhost", help="Postgres database host")
    service_create_parser.add_argument("--db-port", default=5432, type=int, help="Postgres database port")
    service_create_parser.add_argument("-u", "--db-user", default="root", help="Postgres database user")
    service_create_parser.add_argument("-p", "--db-pass", default="root", help="Postgres database password")

    service_create_parser.set_defaults(func_str="service.create.main")


    ##### PARSER FOR taeg service start #####
    service_start_parser = service_subparsers.add_parser("start", description="Start an taeg-service instance")
    service_start_parser.add_argument("-c", "--config", help="Path to config file")
    service_start_parser.add_argument("-e", "--endpoint", help="Address of this VM including port")
    service_start_parser.add_argument("--port", type=int, default=80, help="Port for server to listen on")
    service_start_parser.add_argument("-k", "--google-key", help="Google OAuth key; use environment variable if not specified")
    service_start_parser.add_argument("-s", "--google-secret", help="Google OAuth secret; use environment variable if not specified")
    service_start_parser.add_argument("--db-host", default="localhost", help="Postgres database host")
    service_start_parser.add_argument("--db-port", default=5432, type=int, help="Postgres database port")
    service_start_parser.add_argument("-u", "--db-user", default="root", help="Postgres database user")
    service_start_parser.add_argument("-p", "--db-pass", default="root", help="Postgres database password")
    service_start_parser.add_argument("-l", "--rate-limit", default=120, type=int, help="Rate limit for submissions in seconds")

    service_start_parser.set_defaults(func_str="service.start.main")


    return parser



"""
Utilities for Taeg Grade
"""

import tempfile
import tarfile
import os
import pandas as pd

from contextlib import contextmanager

@contextmanager
def simple_tar(path):
    """
    Context manager that takes a file at ``path`` and creates a temporary tar archive from which the 
    bytes in the file can be read. Yields the file object with the pointer set to the beginning of the
    file. Used for adding files to Docker containers through the Docker Python SDK. Closes and deletes
    the temporary file after the context is closed.

    Args:
        path (``str``): path to the desired file

    Yields:
        ``tempfile.NamedTemporaryFile``: the file with the tar archive written to it
    """
    f = tempfile.NamedTemporaryFile()
    t = tarfile.open(mode="w", fileobj=f)

    path = os.path.abspath(path)
    t.add(path, arcname=os.path.basename(path))

    t.close()
    f.seek(0)

    yield f

    f.close()

@contextmanager
def get_container_file(container, path):
    """
    Retrieves a file at ``path`` from a Docker container ``container``. Reads the bytes of this file
    as a tar archive from the container and writes these bytes to a temporary file. Extracts the single
    member of the tar archive from the temporary file and writes the bytes to another temporary file.
    Yields the file object with the pointer set to the beginning of the file. Closes and deletes the
    temporary files after the context is closed.

    Args:
        container (``docker.models.Container``): the Docker container object
        path (``str``): the path to the file in the container

    Yields:
        ``tempfile.NamedTemporaryFile``: the open temporary file with the extracted contents
    """
    tarf = tempfile.NamedTemporaryFile()
    f = tempfile.NamedTemporaryFile()

    bits, _ = container.get_archive(path)
    for chunk in bits:
        tarf.write(chunk)

    tarf.seek(0)

    tar = tarfile.open(mode="r", fileobj=tarf)

    members = tar.getmembers()
    assert len(members) == 1, "Too many members to extract from container"
    file_contents = tar.extractfile(members[0])
    
    f.write(file_contents.read())
    tar.close()
    tarf.close()
    
    f.seek(0)

    yield f

    f.close()

def list_files(path):
    """
    Returns a list of all non-hidden files in a directory
    
    Args:
        path (``str``): path to a directory
    
    Returns:
        ``list`` of ``str``: list of filenames in the given directory

    """
    return [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file)) and file[0] != "."]

def merge_csv(dataframes):
    """
    Merges dataframes along the vertical axis
    
    Args:
        dataframes (``list`` of ``pandas.core.frame.DataFrame``): list of dataframes with same columns
    
    Returns:
        ``pandas.core.frame.DataFrame``: A merged dataframe resulting from 'stacking' all input dataframes

    """
    final_dataframe = pd.concat(dataframes, axis=0, join='inner').sort_index()
    return final_dataframe



"""
Rmarkdown adapter for Taeg Assign
"""

import os
import copy
import pathlib
import nbformat

from .solutions import is_markdown_solution_cell
from .tests import gen_test_cell, is_test_cell
from .utils import rmd_to_cells, Cell, collapse_empty_cells

from ..assignment import is_assignment_cell, read_assignment_metadata
from ..questions import is_question_cell, read_question_metadata, gen_question_cell
from ..r_adapter.tests import read_test
from ..solutions import has_seed
from ..tests import any_public_tests
from ..utils import is_markdown_cell, EmptyCellException

def transform_notebook(rmd_string, assignment, args):
    """
    Converts a master notebook to an Taeg-formatted solutions notebook, parsing test cells into
    dictionaries ready to be written as OK test files.

    Args:
        rmd_string (``str``): the Rmd document as a string
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
        args (``argparse.Namespace``): parsed command line arguments

    Returns:
        ``tuple(nbformat.NotebookNode, dict)``: the transformed notebook and a dictionary mapping 
            test names to their parsed contents

    """
    cells = rmd_to_cells(rmd_string)
    transformed_cells, test_files = get_transformed_cells(cells, assignment)
    collapse_empty_cells(transformed_cells)

    # if assignment.init_cell and not args.no_init_cell and assignment.is_python:
    #     transformed_cells = [gen_init_cell()] + transformed_cells

    # if assignment.check_all_cell and not args.no_check_all and assignment.is_python:
    #     transformed_cells += gen_check_all_cell()
    
    # if assignment.export_cell and not args.no_export_cell and assignment.is_python:
    #     export_cell = assignment.export_cell
    #     if export_cell is True:
    #         export_cell = {}

    #     transformed_cells += gen_export_cells(
    #         export_cell.get('instructions', ''), 
    #         assignment,
    #         pdf = export_cell.get('pdf', True),
    #         filtering = export_cell.get('filtering', True)
    #     )

    # transformed_nb = copy.deepcopy(nb)
    # transformed_nb['cells'] = transformed_cells

    transformed_rmd_string = "\n".join([cell.source for cell in transformed_cells])

    return transformed_rmd_string, test_files

def get_transformed_cells(cells, assignment):
    """
    Takes in a list of cells from the master notebook and returns a list of cells for the solutions
    notebook. Replaces test cells with a cell calling ``taeg.Notebook.check``, inserts Markdown
    response cells for manual questions with Markdown solutions, and comments out question metadata 
    in question cells, among other things.

    Args:
        cells (``list`` of ``nbformat.NotebookNode``): original code cells
        assignment (``taeg.assign.assignment.Assignment``): the assignment configurations
    
    Returns:
        ``tuple(list, dict)``: list of cleaned notebook cells and a dictionary mapping test names to 
            their parsed contents
    """
    # if assignment.is_r:
    #     from .r_adapter.tests import read_test, gen_test_cell
    # else:
    #     from .tests import read_test, gen_test_cell
    
    transformed_cells, test_files = [], {}
    question_metadata, test_cases, processed_solution, md_has_prompt = {}, [], False, False
    need_close_export, no_solution, in_collapse_block = False, False, False

    for cell in cells:

        if not cell.source.strip():
            transformed_cells.append(cell)
            continue

        if has_seed(cell):
            assignment.seed_required = True            

        # this is the prompt cell or if a manual question then the solution cell
        if question_metadata and not processed_solution:
            assert not is_question_cell(cell), f"Found question cell before end of previous question cell: {cell}"

            # # if this isn't a MD solution cell but in a manual question, it has a Markdown prompt
            # if question_metadata.get('manual', False) and is_markdown_cell(cell) and not is_markdown_solution_cell(cell):
            #     md_has_prompt = True
            #     transformed_cells.append(cell)
            #     continue
            
            # # if this a manual question but not MD solution, it has a code solution cell
            # elif question_metadata.get('manual', False) and not is_markdown_solution_cell(cell):
            #     md_has_prompt = True
            
            # # if there is no prompt, add a prompt cell
            # elif is_markdown_solution_cell(cell) and not md_has_prompt:
            #     transformed_cells.append(gen_markdown_response_cell())

            # if this is a test cell, this question has no response cell for the students, so we don't
            # include it in the output notebook but we need a test file
            if is_test_cell(cell):
                no_solution = True
                test = read_test(cell, question_metadata, assignment, rmd=True)
                test_cases.append(test)

            # elif is_seed_cell(cell):
            #     assignment.seed_required = True
            #     continue

            if not no_solution:
                transformed_cells.append(cell)
            
            processed_solution = True

        # if this is a test cell, parse and add to test_cases
        elif question_metadata and processed_solution and is_test_cell(cell):
            test = read_test(cell, question_metadata, assignment, rmd=True)
            test_cases.append(test)

        # # if this is a solution cell, append. if manual question and no prompt, also append prompt cell
        # elif question_metadata and processed_solution and is_solution_cell(cell):
        #     if is_markdown_solution_cell(cell) and not md_has_prompt:
        #         transformed_cells.append(gen_markdown_response_cell())
        #     transformed_cells.append(cell)

        else:
            # the question is over -- we've seen the question and solution and any tests and now we 
            # need to get ready to process the next question which *could be this cell*
            if question_metadata and processed_solution:

                # create a Notebook.check cell
                if test_cases:
                    check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)

                    # only add to notebook if there's a response cell or if there are public tests
                    if not no_solution or any_public_tests(test_cases):
                        transformed_cells.append(check_cell)

                # # add a cell with <!-- END QUESTION --> if a manually graded question
                # manual = question_metadata.get('manual', False)
                # if manual:
                #     need_close_export = True
                
                # reset vars
                question_metadata, processed_solution, test_cases, md_has_prompt, no_solution = {}, False, [], False, False

            # update assignment config if present; don't add cell to output nb
            if is_assignment_cell(cell):
                assignment.update(read_assignment_metadata(cell))

            # if a question cell, parse metadata, comment out question metadata, and append to nb
            elif is_question_cell(cell):
                question_metadata = read_question_metadata(cell)
                manual = question_metadata.get('manual', False)

                # try:
                #     gen_question_cell(cell, manual, need_close_export)
                #     need_close_export = False
                # except EmptyCellException:
                #     pass

            elif is_markdown_solution_cell(cell):
                # transformed_cells.append(gen_markdown_response_cell())
                transformed_cells.append(cell)

            else:
                assert not is_test_cell(cell), f"Test outside of a question: {cell}"

                # if need_close_export:
                #     if cell['cell_type'] == 'code':
                #         transformed_cells.append(gen_close_export_cell())
                #     else:
                #         cell = add_close_export_to_cell(cell)
                #     need_close_export = False

                transformed_cells.append(cell)

    if test_cases:
        check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)
        if not no_solution or any_public_tests(test_cases):
            transformed_cells.append(check_cell)

    return transformed_cells, test_files



"""
Test result collection via abstract syntax tree transformation
"""

import ast

class CheckCallWrapper(ast.NodeTransformer):
    """
    Visits and replaces nodes in an abstract syntax tree in-place.
    
    Tracks import syntax and instances of ``taeg.Notebook`` in an AST. Wraps calls to 
    ``taeg.Notebook.check`` in calls to ``list.append`` to collect results of execution. Removes calls
    to ``taeg.Notebook.check_all``, `taeg.Notebook.export``, and ``taeg.Notebook.to_pdf``.
    
    Args:
        secret (``str``): random digits string that prevents check function from being modified
    
    Attributes:
        secret (``str``): random digits string that prevents check function from being modified
    """
    TAEG_IMPORT_SYNTAX = "import"
    TAEG_IMPORT_NAME = "taeg"
    TAEG_CLASS_NAME = "Notebook"
    TAEG_INSTANCE_NAME = "grader"

    def __init__(self, secret):
        self.secret = secret

    def check_node_constructor(self, expression):
        """
        Creates node that wraps expression in a list (``check_results_XX``) append call
        
        Args:
            expression (``ast.Name``): name for check function

        Returns:
            ``ast.Call``: function call object from calling check

        """
        args = [expression]
        func = ast.Attribute(
            attr='append',
            value=ast.Name(id='check_results_{}'.format(self.secret), ctx=ast.Load()),
            ctx=ast.Load(),
            keywords=[]
        )
        return ast.Call(func=func, args=args, keywords=[])

    def visit_ImportFrom(self, node):
        """
        Visits ``from ... import ...`` nodes and tracks the import syntax and alias of ``taeg.Notebook``

        Args:
            node (``ast.ImportFrom``): the import node

        Returns:
            ``ast.ImportFrom``: the original node
        """
        if node.module == "taeg" and "Notebook" in [n.name for n in node.names]:
            type(self).TAEG_IMPORT_SYNTAX = "from"
            nb_asname = [n.asname for n in node.names if n.name == "Notebook"][0]
            if nb_asname is not None:
                type(self).TAEG_CLASS_NAME = nb_asname
        return node

    def visit_Import(self, node):
        """
        Visits ``import ...`` nodes and tracks the import syntax and alias of ``taeg``

        Args:
            node (``ast.Import``): the import node

        Returns:
            ``ast.Import``: the original node
        """
        if "taeg" in [n.name for n in node.names]:
            type(self).TAEG_IMPORT_SYNTAX = "import"
            taeg_asname = [n.asname for n in node.names if n.name == "taeg"][0]
            if taeg_asname is not None:
                type(self).TAEG_IMPORT_NAME = taeg_asname
        return node

    def visit_Assign(self, node):
        """
        Visits assignment nodes to determine the name assigned to the instance of ``taeg.Notebook``
        created.

        Args:
            node (``ast.Assign``): the assignment node

        Returns:
            ``ast.Assign``: the original node
        """
        if isinstance(node.value, ast.Call):
            if isinstance(node.value.func, ast.Attribute) and type(self).TAEG_IMPORT_SYNTAX == "import":
                if node.value.func.attr == "Notebook" and isinstance(node.value.func.value, ast.Name):
                    if node.value.func.value.id == type(self).TAEG_IMPORT_NAME:
                        assert len(node.targets) == 1, "error parsing taeg.Notebook instance in ast"
                        type(self).TAEG_INSTANCE_NAME = node.targets[0].id
            elif isinstance(node.value.func, ast.Name) and type(self).TAEG_IMPORT_SYNTAX == "from":
                if node.value.func.id == type(self).TAEG_CLASS_NAME:
                    assert len(node.targets) == 1, "error parsing taeg.Notebook instance in ast"
                    type(self).TAEG_INSTANCE_NAME = node.targets[0].id
        return node

    def visit_Expr(self, node):
        """
        Visits expression nodes and removes them if they are calls to ``taeg.Notebook.check_all``,
        ``taeg.Notebook.export``, or ``taeg.Notebook.to_pdf`` or wraps them using 
        ``CheckCallWrapper.check_node_constructor`` if they are calls to ``taeg.Notebook.check``.

        Args:
            node (``ast.Expr``): the expression node

        Returns:
            ``ast.Expr``: the transformed node
            ``None``: if the node was a removed call
        """
        if isinstance(node.value, ast.Call):
            call_node = node.value
            if isinstance(call_node.func, ast.Attribute):
                if isinstance(call_node.func.value, ast.Name) and call_node.func.value.id == type(self).TAEG_INSTANCE_NAME:
                    if call_node.func.attr in ["check_all", "export", "to_pdf"]:
                        return None
                    elif call_node.func.attr == "check":
                        node.value = self.check_node_constructor(call_node)
        return node



"""
IPython notebook API for Taeg Check
"""

import inspect
import requests
import json
import os
import re
import zipfile
import pickle
import time
import datetime as dt

from getpass import getpass
from glob import glob
from urllib.parse import urljoin
from IPython import get_ipython
from IPython.display import display, HTML, Javascript

from .logs import LogEntry, EventType, Log
from ..execute import check
from ..export import export_notebook
# from .utils import wait_for_save

_API_KEY = None
_TAEG_STATE_FILENAME = ".TAEG_STATE"
_TAEG_LOG_FILENAME = ".TAEG_LOG"
_SHELVE = False

class TestsDisplay:
    """
    Class for stitching together TestCollectionResults objects and displaying them in HTML and plaintext

    Args:
        results (``list`` of ``tuple(str, TestCollectionResults)``): the test names and results for each test
            to be displayed
    """
    def __init__(self, results):
        self.test_names = [r[0] for r in results]
        self.results = [r[1] for r in results]

    def __repr__(self):
        ret = ""
        for name, result in zip(self.test_names, self.results):
            ret += f"{name}:\n{repr(result)}\n\n"
        return ret

    def _repr_html_(self):
        ret = ""
        for name, result in zip(self.test_names, self.results):
            ret += f"<p><strong>{name}:</strong></p>\n{result._repr_html_()}\n\n"
        return ret



class Notebook:
    """
    Notebook class for in-notebook autograding

    Args:
        test_dir (``str``, optional): path to tests directory
    """

    def __init__(self, test_dir="./tests"):
        try:
            global _API_KEY, _SHELVE
            # assert os.path.isdir(test_dir), "{} is not a directory".format(test_dir)

            self._path = test_dir
            self._service_enabled = False
            self._notebook = None


            # assume using taeg service if there is a .taeg file
            taeg_configs = glob("*.taeg")
            if taeg_configs:
                # check that there is only 1 config
                assert len(taeg_configs) == 1, "More than 1 taeg config file found"

                # load in config file
                with open(taeg_configs[0]) as f:
                    self._config = json.load(f)

                _SHELVE = self._config.get("save_environment", False)
                self._service_enabled = "endpoint" in self._config
                self._ignore_modules = self._config.get("ignore_modules", [])
                self._vars_to_store = self._config.get("variables", None)



                # if "notebook" not in self._config:
                #     assert len(glob("*.ipynb")) == 1, "Notebook not specified in taeg config file"
                #     self._notebook = glob("*.ipynb")[0]

                # else:
                self._notebook = self._config["notebook"]

                if self._service_enabled:
                    # check that config file has required info
                    assert all([key in self._config for key in ["endpoint", "assignment_id", "class_id"]]), "Taeg config file missing required information"

                    if "auth" not in self._config:
                        self._config["auth"] = "google"

                    self._google_auth_url = urljoin(self._config["endpoint"], "auth/google")
                    self._default_auth_url = urljoin(self._config["endpoint"], "auth")
                    self._submit_url = urljoin(self._config["endpoint"], "submit")

                    self._auth()

        except Exception as e:
            self._log_event(EventType.INIT, success=False, error=e)
            raise e
        else:
            self._log_event(EventType.INIT)

    # TODO: cut out personal auth?
    def _auth(self):
        """
        Asks student to authenticate with an Taeg Service instance if Taeg Service is configured
        for this notebook

        Raises:
            ``AssertionError``: if Taeg Service is not enabled or an invalid auth provider is indicated
        """
        try:
            global _API_KEY
            assert self._service_enabled, 'notebook not configured for taeg service'
            assert self._config["auth"] in ["google", "default"], "invalid auth provider"

            if _API_KEY is not None:
                self._api_key = _API_KEY
                return

            # have users authenticate with OAuth
            if self._config["auth"] == "google":

                    # send them to google login page
                    display(HTML(f"""
                    <p>Please <a href="{self._google_auth_url}" target="_blank">log in</a> to Taeg Service
                    and enter your API key below.</p>
                    """))

                    self._api_key = input()
            # else have them auth with default auth
            else:
                print("Please enter a username and password.")
                username = input("Username: ")
                password = getpass("Password: ")

                # in-notebook auth
                response = requests.get(url=self._default_auth_url, params={"username":username, "password":password})
                self._api_key = response.content.decode("utf-8")
                # print("Your API Key is {}\n".format())
                # print("Paste this in and hit enter")
                # self._api_key = input()

            # store API key so we don't re-auth every time
            _API_KEY = self._api_key

        except Exception as e:
            self._log_event(EventType.AUTH, success=False, error=e)
            raise e
        else:
            self._log_event(EventType.AUTH)


    def _log_event(self, event_type, results=[], question=None, success=True, error=None, shelve_env={}):
        """
        Logs an event

        Args:
            event_type (``taeg.logs.EventType``): the type of event
            results (``list`` of ``taeg.test_files.abstract_test.TestCollectionResults``, optional):
                the results of any checks recorded by the entry
            question (``str``, optional): the question name for this check
            success (``bool``, optional): whether the operation was successful
            error (``Exception``, optional): the exception thrown by the operation, if applicable
        """
        entry = LogEntry(
            event_type,
            results=results,
            question=question,
            success=success,
            error=error
        )

        if _SHELVE and event_type == EventType.CHECK:
            entry.shelve(
                shelve_env,
                delete=True,
                filename=_TAEG_LOG_FILENAME,
                ignore_modules=self._ignore_modules,
                variables=self._vars_to_store
            )

        entry.flush_to_file(_TAEG_LOG_FILENAME)

    def check(self, question, global_env=None):
        """
        Runs tests for a specific question against a global environment. If no global environment
        is provided, the test is run against the calling frame's environment.

        Args:
            question (``str``): name of question being graded
            global_env (``dict``, optional): global environment resulting from execution of a single
                notebook

        Returns:
            ``taeg.test_files.abstract_test.TestCollectionResults``: the grade for the question
        """
        try:
            test_path = os.path.join(self._path, question + ".py")

            # ensure that desired test exists
            assert os.path.isfile(test_path), "Test {} does not exist".format(question)

            # pass the correct global environment
            if global_env is None:
                global_env = inspect.currentframe().f_back.f_globals

            # run the check
            result = check(test_path, global_env)

        except Exception as e:
            self._log_event(EventType.CHECK, question=question, success=False, error=e, shelve_env=global_env)
            raise e
        else:
            self._log_event(EventType.CHECK, [result], question=question, shelve_env=global_env)

        return result


    # @staticmethod
    def to_pdf(self, nb_path=None, filtering=True, pagebreaks=True, display_link=True):
        """
        Exports a notebook to a PDF. ``filter_type`` can be ``"html"`` or ``"tags"`` if filtering by
        HTML comments or cell tags, respectively.

        Args:
            nb_path (``str``): Path to iPython notebook we want to export
            filtering (``bool``, optional): Set true if only exporting a subset of nb cells to PDF
            pagebreaks (``bool``, optional): If true, pagebreaks are included between questions
            display_link (``bool``, optional): Whether or not to display a download link
        """
        # self._save_notebook()
        try:
            if nb_path is None and self._notebook is not None:
                nb_path = self._notebook

            elif nb_path is None and glob("*.ipynb"):
                notebooks = glob("*.ipynb")
                assert len(notebooks) == 1, "nb_path not specified and > 1 notebook in working directory"
                nb_path = notebooks[0]

            elif nb_path is None:
                raise ValueError("nb_path is None and no taeg-service config is available")

            # convert(nb_path, filtering=filtering, filter_type=filter_type)
            export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)

            if display_link:
                # create and display output HTML
                out_html = """
                <p>Your file has been exported. Download it by right-clicking
                <a href="{}" target="_blank">here</a> and selecting <strong>Save Link As</strong>.
                """.format(nb_path[:-5] + "pdf")

                display(HTML(out_html))

        except Exception as e:
            self._log_event(EventType.TO_PDF, success=False, error=e)
            raise e
        else:
            self._log_event(EventType.TO_PDF)


    def export(self, nb_path=None, export_path=None, pdf=True, filtering=True, pagebreaks=True, files=[], display_link=True):
        """
        Exports a submission to a zipfile. Creates a submission zipfile from a notebook at ``nb_path``,
        optionally including a PDF export of the notebook and any files in ``files``.

        Args:
            nb_path (``str``): path to notebook we want to export
            export_path (``str``, optional): path at which to write zipfile
            pdf (``bool``, optional): whether a PDF should be included
            filtering (``bool``, optional): whether the PDF should be filtered; ignored if ``pdf`` is
                ``False``
            pagebreaks (``bool``, optional): whether pagebreaks should be included between questions
            files (``list`` of ``str``, optional): paths to other files to include in the zipfile
            display_link (``bool``, optional): whether or not to display a download link
        """
        self._log_event(EventType.BEGIN_EXPORT)
        # self._save_notebook()

        try:
            if nb_path is None and self._notebook is not None:
                nb_path = self._notebook

            elif nb_path is None and glob("*.ipynb"):
                notebooks = glob("*.ipynb")
                assert len(notebooks) == 1, "nb_path not specified and > 1 notebook in working directory"
                nb_path = notebooks[0]

            elif nb_path is None:
                raise ValueError("nb_path is None and no taeg-service config is available")

            try:
                with open(nb_path) as f:
                    assert len(f.read().strip()) > 0, \
                        f"Notebook {nb_path} is empty. Please save and checkpoint your notebook and rerun this cell."

            except UnicodeDecodeError:
                with open(nb_path, "r", encoding="utf-8") as f:
                    assert len(f.read().strip()) > 0, \
                        f"Notebook {nb_path} is empty. Please save and checkpoint your notebook and rerun this cell."

            if export_path is None:
                zip_path = ".".join(nb_path.split(".")[:-1]) + ".zip"
            else:
                zip_path = export_path

            zf = zipfile.ZipFile(zip_path, mode="w")
            zf.write(nb_path)

            if pdf:
                pdf_path = ".".join(nb_path.split(".")[:-1]) + ".pdf"
                # convert(nb_path, filtering=filtering, filter_type=filter_type)
                export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
                zf.write(pdf_path)

            if os.path.isfile(_TAEG_LOG_FILENAME):
                zf.write(_TAEG_LOG_FILENAME)

            if glob("*.taeg"):
                assert len(glob("*.taeg")) == 1, "Too many .taeg files (max 1 allowed)"
                zf.write(glob("*.taeg")[0])

            for file in files:
                zf.write(file)

            zf.close()

            if display_link:
                # create and display output HTML
                out_html = """
                <p>Your submission has been exported. Click <a href="{}" target="_blank">here</a>
                to download the zip file.</p>
                """.format(zip_path)

                display(HTML(out_html))

        except Exception as e:
            self._log_event(EventType.END_EXPORT, success=False, error=e)
            raise e
        else:
            self._log_event(EventType.END_EXPORT)

    def check_all(self):
        """
        Runs all tests on this notebook. Tests are run against the current global environment, so any
        tests with variable name collisions will fail.
        """
        # TODO: this should use functions in execute.py to run tests in-sequence so that variable
        # name collisions are accounted for
        self._log_event(EventType.BEGIN_CHECK_ALL)

        try:
            tests = glob(os.path.join(self._path, "*.py"))
            global_env = inspect.currentframe().f_back.f_globals
            results = []
            if not _SHELVE:
                for file in sorted(tests):
                    if "__init__.py" not in file:
                        test_name = os.path.split(file)[1][:-3]
                        result = self.check(test_name, global_env)
                        results.append((test_name, result))
            else:
                log = Log.from_file(_TAEG_LOG_FILENAME, ascending=False)
                for file in sorted(tests):
                    if "__init__.py" not in file:
                        test_name = os.path.splitext(os.path.split(file)[1])[0]

                        entry = log.get_question_entry(test_name)
                        env = entry.unshelve()
                        global_env.update(env)
                        del locals()["env"]

                        result = self.check(test_name, global_env)
                        results.append((test_name, result))

        except Exception as e:
            self._log_event(EventType.END_CHECK_ALL, success=False, error=e)
            raise e
        else:
            self._log_event(EventType.END_CHECK_ALL)

        return TestsDisplay(results)

    def submit(self):
        """
        Submits this notebook to an Taeg Service instance if Taeg Service is configured

        Raises:
            ``AssertionError``: if this notebook is not configured for Taeg Service
        """
        assert self._service_enabled, 'notebook not configured for taeg service'

        try:
            if not hasattr(self, '_api_key'):
                self._auth()


            notebook_path = os.path.join(os.getcwd(), self._notebook)

            assert os.path.exists(notebook_path) and os.path.isfile(notebook_path), \
            "Could not find notebook: {}".format(self._notebook)

            with open(notebook_path) as f:
                notebook_data = json.load(f)

            notebook_data["metadata"]["assignment_id"] = self._config["assignment_id"]
            notebook_data["metadata"]["class_id"] = self._config["class_id"]

            print("Submitting notebook to server...")

            response = requests.post(self._submit_url, json.dumps({
                "api_key": self._api_key,
                "nb": notebook_data,
            }))

        except Exception as e:
            self._log_event(EventType.SUBMIT, success=False, error=e)
            raise e
        else:
            self._log_event(EventType.SUBMIT)



"""
Support for OK-formatted test files
"""

import os
import io
import doctest
import warnings
import pathlib

from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent

from .abstract_test import TestFile
from ..utils import hide_outputs

def run_doctest(name, doctest_string, global_environment):
    """
    Run a single test with given ``global_environment``. Returns ``(True, '')`` if the doctest passes. 
    Returns ``(False, failure_message)`` if the doctest fails.

    Args:
        name (``str``): name of doctest
        doctest_string (``str``): doctest in string form
        global_environment (``dict``): global environment resulting from the execution of a python 
            script/notebook
    
    Returns:
        ``tuple`` of (``bool``, ``str``): results from running the test
    """
    examples = doctest.DocTestParser().parse(
        doctest_string,
        name
    )
    test = doctest.DocTest(
        [e for e in examples if isinstance(e, doctest.Example)],
        global_environment,
        name,
        None,
        None,
        doctest_string
    )

    doctestrunner = doctest.DocTestRunner(verbose=True)

    runresults = io.StringIO()
    with redirect_stdout(runresults), redirect_stderr(runresults), hide_outputs():
        doctestrunner.run(test, clear_globs=False)
    with open(os.devnull, 'w') as f, redirect_stderr(f), redirect_stdout(f):
        result = doctestrunner.summarize(verbose=True)
    # An individual test can only pass or fail
    if result.failed == 0:
        return (True, '')
    else:
        return False, runresults.getvalue()

class OKTestFile(TestFile):
    """
    A single test (set of doctests) for Taeg.
    
    Instances of this class are callable. When called, it takes a ``global_environment`` dict, and returns 
    an ``taeg.ok_parser.OKTestsResult`` object. We only take a global_environment, *not* a 
    ``local_environment``. This makes tests not useful inside functions, methods or other scopes with 
    local variables. This is a limitation of doctest, so we roll with it.

    The last 2 attributes (``passed``, ``failed_test``) are set after calling ``run()``.

    Args:
        name (``str``): name of test
        tests (``list`` of ``str``): ;ist of docstring tests to be run
        hiddens (``list`` of ``bool``): visibility of each tests in ``tests``
        value (``int``, optional): point value of this test, defaults to 1
        hidden (``bool``, optional): wheter this test is hidden
    """

    def run(self, global_environment):
        """Runs tests on a given ``global_environment``
        
        Arguments:
            ``global_environment`` (``dict``): result of executing a Python notebook/script
        
        Returns:
            ``tuple`` of (``bool``, ``taeg.ok_parser.OKTest``): whether the test passed and a pointer 
                to the current ``taeg.ok_parser.OKTest`` object
        """
        for i, t in enumerate(self.public_tests + self.hidden_tests):
            passed, result = run_doctest(self.name + ' ' + str(i), t, global_environment)
            if not passed:
                self.passed = False
                self.failed_test = t
                self.failed_test_hidden = i >= len(self.public_tests)
                self.result = result
                return False, self
        self.passed = True
        return True, self

    @classmethod
    def from_file(cls, path):
        """
        Parse an ok test file & return an ``OKTest``

        Args:
            path (``str``): path to ok test file

        Returns:
            ``taeg.ok_parser.OKTest``: new ``OKTest`` object created from the given file
        """
        # ok test files are python files, with a global 'test' defined
        test_globals = {}
        with open(path) as f:
            exec(f.read(), test_globals)

        test_spec = test_globals['test']

        # We only support a subset of these tests, so let's validate!

        # Make sure there is a name
        assert 'name' in test_spec

        # Do not support multiple suites in the same file
        assert len(test_spec['suites']) == 1

        test_suite = test_spec['suites'][0]

        # Only support doctest. I am unsure if other tests are implemented
        assert test_suite.get('type', 'doctest') == 'doctest'

        # Not setup and teardown supported
        assert not bool(test_suite.get('setup'))
        assert not bool(test_suite.get('teardown'))

        if 'hidden' in test_spec:
            warnings.warn(
                "The global 'hidden' key of ok-tests is deprecated since v1.0.0. "
                "This key will be ignored.", 
                FutureWarning
            )

        tests = []
        hiddens = []

        for _, test_case in enumerate(test_spec['suites'][0]['cases']):
            tests.append(dedent(test_case['code']))
            hiddens.append(test_case.get('hidden', True))

        # convert path into PurePosixPath for test name
        name = str(pathlib.Path(path).as_posix())

        return cls(name, tests, hiddens, test_spec.get('points', 1), test_spec.get('hidden', True))



"""
Logging for Taeg Check
"""

import os
import pickle
import types
import dill
import tempfile
import datetime as dt
import numpy as np

from enum import Enum, auto
from glob import glob

class QuestionNotInLogException(Exception):
    """Exception that indicates that a specific question was not found in any entry in the log"""


class EventType(Enum):
    """Enum of event types for log entries
    
    Attributes:
        AUTH: an auth event
        BEGIN_CHECK_ALL: beginning of a check-all call
        BEGIN_EXPORT: beginning of an assignment export
        CHECK: a check of a single question
        END_CHECK_ALL: ending of a check-all call
        END_EXPORT: ending of an assignment export
        INIT: initialization of an ``taeg.Notebook`` object
        SUBMIT: submission of an assignment
        TO_PDF: PDF export of a notebook
    """

    AUTH = auto()
    BEGIN_CHECK_ALL = auto()
    BEGIN_EXPORT = auto()
    CHECK = auto()
    END_CHECK_ALL = auto()
    END_EXPORT = auto()
    INIT = auto()
    SUBMIT = auto()
    TO_PDF = auto()


class LogEntry:
    """
    An entry in Taeg's log. Tracks event type, grading results, success of operation, and errors
    thrown.

    Args:
        event_type (``EventType``): the type of event for this entry
        results (``list`` of ``taeg.test_files.abstract_test.TestCollectionResults``, optional): the 
            results of grading if this is an ``EventType.CHECK`` record
        question (``str``, optional): the question name for an ``EventType.CHECK`` record
        success (``bool``, optional): whether the operation was successful
        error (``Exception``, optional): an error thrown by the process being logged if any

    Attributes:
        event_type (``EventType``): the entry type
        shelf (``bytes``): a pickled environment stored as a bytes string
        unshelved (``list`` of ``str``): a list of variable names that were unable to be pickled during
            shelving
        results (``list`` of ``taeg.test_files.abstract_test.TestCollectionResults``): grading results 
            if this is an ``EventType.CHECK`` entry
        question (``str``): question name if this is a check entry
        success (``bool``): whether the operation tracked by this entry was successful
        error (``Exception``): an error thrown by the tracked process if applicable
        timestamp (``datetime.datetime``): timestamp of event in UTC
    """

    def __init__(self, event_type, shelf=None, unshelved=[], results=[], question=None, success=True, error=None):
        assert event_type in EventType, "Invalid event type"
        self.event_type = event_type
        self.shelf = shelf
        self.unshelved = []
        self.results = results
        self.question = question
        self.timestamp = dt.datetime.utcnow()
        self.success = success
        self.error = error

    def __repr__(self):
        if self.question:
            return "taeg.logs.LogEntry(event_type={}, question={}, success={}, timestamp={})".format(
                self.event_type, self.question, self.success, self.timestamp.isoformat()
            )

        return "taeg.logs.LogEntry(event_type={}, success={}, timestamp={})".format(
            self.event_type, self.success, self.timestamp.isoformat()
        )

    def get_results(self):
        """
        Get the results stored in this log entry
        
        Returns:
            ``list`` of ``taeg.test_files.abstract_test.TestCollectionResults``: the results at this 
                entry if this is an ``EventType.CHECK`` record
        """
        assert self.event_type is EventType.CHECK, "this record type has no results"
        if isinstance(self.results, list):
            return self.results[0]
        return self.results

    def get_score_perc(self):
        """
        Returns the percentage score for the results of this entry

        Returns:
            ``float``: the percentage score
        """
        return self.get_results().grade

    def raise_error(self):
        """
        Raises the error stored in this entry

        Raises:
            ``Exception``: the error stored at this entry, if present
        """
        if self.error is not None:
            raise self.error

    def flush_to_file(self, filename):
        """
        Appends this log entry (pickled) to a file
        
        Args:
            filename (``str``): the path to the file to append this entry
        """
        try:
            file = open(filename, "ab+")
            pickle.dump(self, file)

        except OSError:
            raise Exception(
                "Could not create the log file as the file system is read-only. Please contact your "
                "instructor before continuing on this assignment."
            )

        finally:
            if "file" in locals():
                file.close()

    def shelve(self, env, delete=False, filename=None, ignore_modules=[], variables=None):
        """
        Stores an environment ``env`` in this log entry using dill as a ``bytes`` object in this entry
        as the ``shelf`` attribute. Writes names of any variables in ``env`` that are not stored to
        the ``unshelved`` attribute.

        If ``delete`` is ``True``, old environments in the log at ``filename`` for this question are
        cleared before writing ``env``. Any module names in ``ignore_modules`` will have their functions
        ignored during pickling. 

        Args:
            env (``dict``): the environment to pickle
            delete (``bool``, optional): whether to delete old environments
            filename (``str``, optional): path to log file; ignored if ``delete`` is ``False``
            ignore_modules (``list`` of ``str``, optional): module names to ignore during pickling
            variables (``dict``, optional): map of variable name to type string indicating **only** 
                variables to include (all variables not in this dictionary will be ignored)

        Returns:
            ``LogEntry``: this entry
        """
        # delete old entry without reading entire log
        if delete:
            assert filename, "old env deletion indicated but no log filename provided"
            try:                
                # copy the existing log to a temporary file so that we can edit the original
                with tempfile.TemporaryFile() as tf:
                    with open(filename, "rb") as f:
                        tf.write(f.read())
                    tf.seek(0)
                    os.system(f"rm -f {filename}")
                    while True:
                        try:
                            entry = pickle.load(tf)

                            if entry.question == self.question and entry.shelf is not None:

                                # only edit variables if it's not provided
                                if variables is None:
                                    variables = list(entry.unshelve().keys())
                                    variables_stored = None
                                else:
                                    variables_stored = list(entry.unshelve().keys())
                                
                                entry.shelf = None

                            entry.flush_to_file(filename)

                            del locals()["entry"]

                        except EOFError:
                            break

            except FileNotFoundError:
                pass

        try:
            if isinstance(variables_stored, list):
                variables = {k : v for k, v in variables.items() if k in variables_stored}
        except UnboundLocalError:
            pass

        shelf_contents, unshelved = LogEntry.shelve_environment(env, variables=variables, ignore_modules=ignore_modules)
        self.shelf = shelf_contents
        self.unshelved = unshelved
        return self
        
    def unshelve(self, global_env={}):
        """
        Parses a ``bytes`` object stored in the ``shelf`` attribute and unpickles the object stored
        there using dill. Updates the ``__globals__`` of any functions in ``shelf`` to include elements
        in the shelf. Optionally includes the env passed in as ``global_env``.

        Args:
            global_env (``dict``, optional): a global env to include in unpickled function globals

        Returns:
            ``dict``: the shelved environment
        """
        assert self.shelf, "no shelf in this entry"

        # read bytes in self.shelf and load with dill
        with tempfile.TemporaryFile() as tf:
            tf.write(self.shelf)
            tf.seek(0)
            shelf = dill.load(tf)
            
        # add the unpickeld env and global_env to all function __globals__
        for k, v in shelf.items():
            if type(v) == types.FunctionType:
                v.__globals__.update(shelf)
                v.__globals__.update(global_env)

        return shelf

    @staticmethod
    def sort_log(log, ascending=True):
        """
        Sorts a list of log entries by timestamp
        
        Args:
            log (``list`` of ``LogEntry``): the log to sort
            ascending (``bool``, optional): whether the log should be sorted in ascending (chronological) 
                order; default ``True``

        Returns:
            ``list`` of ``LogEntry``: the sorted log
        """
        if ascending:
            return list(sorted(log, key = lambda l: l.timestamp))
        return list(sorted(log, key = lambda l: l.timestamp, reverse = True))

    @staticmethod
    def log_from_file(filename, ascending=True):
        """
        Reads a log file and returns a sorted list of the log entries pickled in that file
        
        Args:
            filename (``str``): the path to the log
            ascending (``bool``, optional): whether the log should be sorted in ascending (chronological) 
                order; default ``True``

        Returns:
            ``list`` of ``LogEntry``: the sorted log
        """
        try:
            file = open(filename, "rb")

            log = []
            while True:
                try:
                    log.append(pickle.load(file))
                except EOFError:
                    break

            log = list(sorted(log, key = lambda l: l.timestamp, reverse = not ascending))
            
            return log
            
        finally:
            file.close()

    @staticmethod
    def shelve_environment(env, variables=None, ignore_modules=[]):
        """
        Pickles an environment ``env`` using dill, ignoring any functions whose module is listed in
        ``ignore_modules``. Returns the pickle file contents as a ``bytes`` object and a list of
        variable names that were unable to be shelved/ignored during shelving.

        Args:
            env (``dict``): the environment to shelve
            variables (``dict`` *or* ``list``, optional): a map of variable name to type string indicating 
                **only** variables to include (all variables not in this dictionary will be ignored) 
                or a list of variable names to include regardless of tpye
            ignore_modules (``list`` of ``str``, optional): the module names to igonre

        Returns:
            ``tuple`` of (``bytes``, ``list`` of ``str``): the pickled environment and list of unshelved
                variable names.
        """
        from .notebook import Notebook
        unshelved = []
        filtered_env = {}
        for k, v in env.items():

            # don't store modules or taeg.Notebook instances
            if type(v) == types.ModuleType or type(v) == Notebook:
                unshelved.append(k)
            
            # ignore any functions whose __module__ is in ignore_modules
            elif type(v) == types.FunctionType and v.__module__ in ignore_modules:
                unshelved.append(k)
            
            # ensure object is pickleable by attempting dump and if so add to filtered_env
            else:
                try:
                    dill.dumps(v)

                    # only store variable names in variables that have the correct type
                    if (variables and k in variables) or not variables:
                        full_type = type(v).__module__ + "." + type(v).__name__

                        if (isinstance(variables, dict) and full_type == variables[k]) or \
                            isinstance(variables, list) or not variables:
                            filtered_env[k] = v

                        else:
                            unshelved.append(k)

                    else:
                        unshelved.append(k)
                        
                except:
                    unshelved.append(k)

        # dump filtered_env to a temporary file and then return the bytes and unshelved list
        with tempfile.TemporaryFile() as tf:
            dill.dump(filtered_env, tf)
            tf.seek(0)
            shelf_contents = tf.read()
            
        return shelf_contents, unshelved


class Log:
    """
    A class for reading and interacting with a log. Allows you to iterate over the entries in the log 
    and supports integer indexing. *Does not support editing the log file.*

    Args:
        entries (``list`` of ``LogEntry``): the list of entries for this log
        ascending (``bool``, optional): whether the log is sorted in ascending (chronological) order;
            default ``True``

    Attributes:
        entries (``list`` of ``LogEntry``): the list of log entries in this log
        ascending (``bool``): whether ``entries`` is sorted chronologically; ``False`` indicates reverse-
            chronological order
    """

    def __init__(self, entries, ascending=True):
        self.entries = entries
        self.ascending = ascending

    def __repr__(self):
        return "taeg.logs.Log([\n  {}\n])".format(",\n  ".join([repr(e) for e in self.entries]))

    def __getitem__(self, idx):
        return self.entries[idx]

    def __iter__(self):
        return iter(self.entries)

    def question_iterator(self):
        """
        Returns an iterator over the most recent entries for each question.

        Returns:
            ``QuestionLogIterator``: the iterator
        """
        return QuestionLogIterator(self)

    def sort(self, ascending=True):
        """
        Sorts this logs entries by timestmap using ``LogEntry.sort_log``.

        Args:
            ascending (``bool``, optional): whether to sort the log chronologically; defaults to ``True``
        """
        self.entries = LogEntry.sort_log(self.entries, ascending=ascending)
        # self.entries = LogEntry.sort_log(self.entries)
        self.ascending = ascending

    def get_questions(self):
        """
        Returns a sorted list of all question names that have entries in this log.

        Returns:
            ``list`` of ``str``: the questions in this log
        """
        all_questions = [entry.question for entry in self.entries if entry.event_type == EventType.CHECK]
        return list(sorted(set(all_questions)))

    @classmethod
    def from_file(cls, filename, ascending=True):
        """
        Loads and sorts a log from a file.

        Args:
            filename (``str``): the path to the log
            ascending (``bool``, optional): whether the log should be sorted in ascending (chronological) 
                order; default ``True``

        Returns:
            ``Log``: the ``Log`` instance created from the file
        """
        return cls(entries=LogEntry.log_from_file(filename, ascending=ascending), ascending=ascending)

    def get_question_entry(self, question):
        """
        Gets the most recent entry corresponding to the question ``question``

        Args:
            question (``str``): the question to get

        Returns:
            ``LogEntry``: the most recent log entry for ``question``
        
        Raises:
            ``QuestionNotInLogException``: if the question is not in the log
        """
        if self.ascending:
            self.entries = LogEntry.sort_log(self.entries, ascending=False)
            self.ascending = False
        for entry in self.entries:
            if entry.question == question:
                return entry
        raise QuestionNotInLogException(f"question {question} is not in the log")

    def get_results(self, question):
        """Gets the most recent grading result for a specified question from this log

        Args:
            question (``str``): the question name to look up

        Returns:
            ``taeg.test_files.abstract_test.TestCollectionResults``: the most recent result for the question

        Raises:
            ``QuestionNotInLogException``: if the question is not found
        """
        return self.get_question_entry(question).get_results()

class QuestionLogIterator:
    """
    An iterator over the most recent entries for each question in the log. Sorts the log when initialized
    and uses `Log.get_questions` to retrieve the list of questions.

    Args:
        log (``Log``): the log over which to iterate

    Attributes:
        log (``Log``): the log being iterated over
        questions (``list`` of ``str``): the list of question names
        curr_idx (``int``): the integer index of the next question in  ``questions``
    """
    def __init__(self, log):
        log.sort(ascending=False)
        log.sort()
        self.log = log
        self.questions = self.log.get_questions()
        self.curr_idx = 0
    
    def __iter__(self):
        return self
    
    def __next__(self):
        if self.curr_idx >= len(self.questions):
            raise StopIteration

        question = self.questions[self.curr_idx]
        entry = self.log.get_question_entry(question)
        self.curr_idx += 1
        return entry
